@@ -1,4 +1,5 @@
-/*
+/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
+ *
  *  arch-tag: Implementation of simple Rhythmbox debugging interface
  *
  *  Copyright (C) 2002 Jorn Baayen
@@ -15,19 +16,21 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
  *
  *  NOTES: log domain hack stolen from nautilus
  *
  */
 
-#include <glib.h>
+#include <unistd.h>
 #include <stdio.h>
 #include <string.h>
 #include <stdarg.h>
 #include <signal.h>
 #include <time.h>
 
+#include <glib.h>
+
 #include "rb-debug.h"
 
 static void log_handler (const char *domain,
@@ -35,7 +38,8 @@ static void log_handler (const char *domain,
 	                 const char *message,
 	                 gpointer data);
 
-static gboolean debugging = FALSE;
+static const char *debug_everything = "everything";
+static const char *debug_match = NULL;
 
 /* Our own funky debugging function, should only be used when something
  * is not going wrong, if something *is* wrong use g_warning.
@@ -51,11 +55,15 @@ rb_debug_real (const char *func,
 	char *str_time;
 	time_t the_time;
 
-	if (debugging == FALSE) return;
+	if (debug_match == NULL || 
+	   (debug_match != debug_everything &&
+	   (strstr (file, debug_match) == NULL) &&
+	   (strstr (func, debug_match) == NULL)))
+		return;
 
 	va_start (args, format);
 
-	vsnprintf (buffer, 1024, format, args);
+	g_vsnprintf (buffer, 1024, format, args);
 	
 	va_end (args);
 
@@ -63,14 +71,20 @@ rb_debug_real (const char *func,
 	str_time = g_new0 (char, 255);
 	strftime (str_time, 254, "%H:%M:%S", localtime (&the_time));
 
-	g_printerr ("[%p] [%s] %s:%d (%s): %s\n", g_thread_self (),
-		    func, file, line, str_time, buffer);
+	g_printerr ("(%s) [%p] [%s] %s:%d: %s\n", str_time, g_thread_self (),
+		    func, file, line, buffer);
 	
 	g_free (str_time);
 }
 
 void
 rb_debug_init (gboolean debug)
+{
+	rb_debug_init_match (debug ? debug_everything : NULL);
+}
+
+void 
+rb_debug_init_match (const char *match)
 {
 	guint i;
 
@@ -94,6 +108,7 @@ rb_debug_init (gboolean debug)
 		"GLib-GObject",
 		"GModule",
 		"GThread",
+		"GStreamer",
 		"Gdk",
 		"Gdk-Pixbuf",
 		"GdkPixbuf",
@@ -119,12 +134,11 @@ rb_debug_init (gboolean debug)
 		"librsvg",
 	};
 
-	debugging = debug;
+	debug_match = match;
 
-	for (i = 0; i < G_N_ELEMENTS (standard_log_domains); i++)
-	{
-		g_log_set_handler (standard_log_domains[i], G_LOG_LEVEL_MASK, log_handler, NULL);
-	}
+	if (debug_match != NULL)
+		for (i = 0; i < G_N_ELEMENTS (standard_log_domains); i++)
+			g_log_set_handler (standard_log_domains[i], G_LOG_LEVEL_MASK, log_handler, NULL);
 
 	rb_debug ("Debugging enabled");
 }
@@ -171,7 +185,7 @@ rb_profiler_new (const char *name)
 {
 	RBProfiler *profiler;
 	
-	if (debugging == FALSE)
+	if (debug_match == NULL)
 		return NULL;
 
 	profiler = g_new0 (RBProfiler, 1);
@@ -186,16 +200,16 @@ rb_profiler_new (const char *name)
 void
 rb_profiler_dump (RBProfiler *profiler)
 {
-	long elapsed;
+	gulong elapsed;
 	double seconds;
 
-	if (debugging == FALSE)
+	if (debug_match == NULL)
 		return;
 	if (profiler == NULL)
 		return;
 
 	seconds = g_timer_elapsed (profiler->timer, &elapsed);
-	
+
 	rb_debug ("PROFILER %s %ld ms (%f s) elapsed", profiler->name, 
 		  elapsed / (G_USEC_PER_SEC / 1000), seconds);
 }
@@ -203,7 +217,7 @@ rb_profiler_dump (RBProfiler *profiler)
 void
 rb_profiler_reset (RBProfiler *profiler)
 {
-	if (debugging == FALSE)
+	if (debug_match == NULL)
 		return;
 	if (profiler == NULL)
 		return;
@@ -214,7 +228,7 @@ rb_profiler_reset (RBProfiler *profiler)
 void
 rb_profiler_free (RBProfiler *profiler)
 {
-	if (debugging == FALSE)
+	if (debug_match == NULL)
 		return;
 	if (profiler == NULL)
 		return;
@@ -223,3 +237,45 @@ rb_profiler_free (RBProfiler *profiler)
 	g_free (profiler->name);
 	g_free (profiler);
 }
+
+/* Profiling */
+
+static int profile_indent;
+
+static void
+profile_add_indent (int indent)
+{
+	profile_indent += indent;
+	if (profile_indent < 0) {
+		g_error ("You screwed up your indentation");
+	}
+}
+
+void
+_rb_profile_log (const char *func,
+		 const char *file,
+		 int         line,
+		 int	     indent,
+		 const char *msg1,
+		 const char *msg2)
+{
+	char *str;
+
+	if (indent < 0) {
+		profile_add_indent (indent);
+	}
+
+	if (profile_indent == 0) {
+		str = g_strdup_printf ("MARK: [%s %s %d] %s %s", file, func, line, msg1 ? msg1 : "", msg2 ? msg2 : "");
+	} else {
+		str = g_strdup_printf ("MARK: %*c [%s %s %d] %s %s", profile_indent - 1, ' ', file, func, line, msg1 ? msg1 : "", msg2 ? msg2 : "");
+	}
+
+	access (str, F_OK);
+
+	g_free (str);
+
+	if (indent > 0) {
+		profile_add_indent (indent);
+	}
+}
