@@ -1,8 +1,8 @@
-/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*- */
-/*
+/* -*- Mode: C; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 8 -*-
+ *
  * arch-tag: Implementation of GStreamer recorder backend
  *
- * Copyright (C) 2004-2005 William Jon McCann <mccann@jhu.edu>
+ * Copyright (C) 2004-2006 William Jon McCann <mccann@jhu.edu>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -16,11 +16,12 @@
  *
  * You should have received a copy of the GNU General Public
  * License along with this program; if not, write to the
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA 02111-1307, USA.
+ * Free Software Foundation, Inc., 51 Franklin St, Fifth Floor,
+ * Boston, MA 02110-1301  USA.
  */
 
-#include <config.h>
+#include "config.h"
+
 #include <unistd.h>
 #include <string.h>
 #include <math.h>
@@ -32,13 +33,22 @@
 
 #include <glib/gi18n.h>
 #include <libgnomevfs/gnome-vfs-utils.h>
-#include <nautilus-burn-recorder.h>
 #include <gst/gst.h>
 #ifdef HAVE_GSTREAMER_0_8
 #include <gst/gconf/gconf.h>
 #include <gst/play/play.h>
 #endif
 
+#include <nautilus-burn-drive.h>
+#include <nautilus-burn-recorder.h>
+#ifndef NAUTILUS_BURN_CHECK_VERSION 	 
+#define NAUTILUS_BURN_CHECK_VERSION(a,b,c) FALSE 	 
+#endif
+
+#if NAUTILUS_BURN_CHECK_VERSION(2,15,3)
+#include <nautilus-burn.h>
+#endif
+
 #include "rb-recorder.h"
 
 #include "rb-debug.h"
@@ -123,8 +133,9 @@ GQuark
 rb_recorder_error_quark (void)
 {
         static GQuark quark = 0;
-        if (!quark)
+        if (! quark) {
                 quark = g_quark_from_static_string ("rb_recorder_error");
+        }
 
         return quark;
 }
@@ -221,10 +232,17 @@ rb_recorder_get_default_drive (void)
         NautilusBurnDrive *drive  = NULL;
         GList             *drives = NULL;
 
+#if NAUTILUS_BURN_CHECK_VERSION(2,15,3)
+        NautilusBurnDriveMonitor *monitor;
+        monitor = nautilus_burn_get_drive_monitor ();
+        drives = nautilus_burn_drive_monitor_get_recorder_drives (monitor);
+#else
         drives = nautilus_burn_drive_get_list (TRUE, FALSE);
+#endif
 
-        if (drives)
+        if (drives) {
                 drive = nautilus_burn_drive_ref ((NautilusBurnDrive*) drives->data);
+        }
 
         g_list_foreach (drives, (GFunc)nautilus_burn_drive_unref, NULL);
         g_list_free (drives);
@@ -264,8 +282,9 @@ rb_recorder_gst_free_pipeline (RBRecorder *recorder)
 {
         rb_debug ("Freeing rb_recorder pipeline");
 
-        if (recorder->priv->pipeline == NULL)
+        if (recorder->priv->pipeline == NULL) {
                 return;
+        }
 
         if (recorder->priv->tick_timeout_id > 0) {
                 g_source_remove (recorder->priv->tick_timeout_id);
@@ -322,8 +341,9 @@ add_track (RBRecorder *recorder,
 
         track->type = NAUTILUS_BURN_RECORDER_TRACK_TYPE_AUDIO;
         track->contents.audio.filename = filename;
-        if (cdtext)
+        if (cdtext) {
                 track->contents.audio.cdtext = g_strdup (cdtext);
+        }
 
         recorder->priv->tracks = g_list_append (recorder->priv->tracks, track);
 
@@ -337,8 +357,9 @@ eos_cb (GstElement *element,
 {
         rb_debug ("EOS");
 
-        if (recorder->priv->pipeline)
+        if (recorder->priv->pipeline) {
                 gst_element_set_state (recorder->priv->pipeline, GST_STATE_NULL);
+        }
 
         g_signal_emit (G_OBJECT (recorder), rb_recorder_signals [EOS], 0);
 }
@@ -352,8 +373,9 @@ error_signal_idle (RBRecorderSignal *signal)
                        signal->error);
 
         /* close if not already closing */
-        if (signal->object->priv->src_uri != NULL)
+        if (signal->object->priv->src_uri != NULL) {
                 rb_recorder_close (signal->object, NULL);
+        }
 
         g_object_unref (signal->object);
         g_error_free (signal->error);
@@ -406,8 +428,9 @@ rb_recorder_gst_signal_error (RBRecorder *recorder,
 		       error);
 
 	/* close if not already closing */
-	if (recorder->priv->src_uri != NULL)
+	if (recorder->priv->src_uri != NULL) {
 		rb_recorder_close (recorder, NULL);
+        }
 
 	g_object_unref (recorder);
 	g_error_free (error);
@@ -470,15 +493,16 @@ rb_recorder_new_pad_cb (GstElement *decodebin,
 	/* Only link audio. */
 	caps = gst_pad_get_caps (pad);
 	str = gst_caps_get_structure (caps, 0);
-	if (!g_strrstr (gst_structure_get_name (str), "audio")) {
+	if (! g_strrstr (gst_structure_get_name (str), "audio")) {
 		gst_caps_unref (caps);
 		return;
 	}
 
 	gst_caps_unref (caps);
 #ifdef HAVE_GSTREAMER_0_10
-	if (gst_pad_link (pad, audio_pad))
+	if (gst_pad_link (pad, audio_pad)) {
 		recorder->priv->got_audio_pad = TRUE;
+        }
 #endif
 }
 
@@ -509,7 +533,7 @@ rb_recorder_construct (RBRecorder *recorder,
          */
 
         recorder->priv->pipeline = gst_pipeline_new ("pipeline");
-        if (!recorder->priv->pipeline) {
+        if (! recorder->priv->pipeline) {
                 g_set_error (error,
                              RB_RECORDER_ERROR,
                              RB_RECORDER_ERROR_GENERAL,
@@ -609,7 +633,7 @@ rb_recorder_construct (RBRecorder *recorder,
                 g_signal_connect_object (G_OBJECT (recorder->priv->pipeline), "eos",
                                          G_CALLBACK (eos_cb), recorder, 0);
 #elif HAVE_GSTREAMER_0_10
-	g_object_set (recorder->priv->capsfilter, "filter-caps",  filtercaps, NULL);
+	g_object_set (recorder->priv->capsfilter, "caps",  filtercaps, NULL);
 
 	gst_element_link_many (recorder->priv->src,
 			       recorder->priv->typefind,
@@ -649,8 +673,10 @@ recorder_track_free (NautilusBurnRecorderTrack *track)
                                               ext - track->contents.audio.filename);
 
                 if (g_file_test (track->contents.audio.filename, G_FILE_TEST_EXISTS)
-                    && unlink (track->contents.audio.filename) != 0)
+                    && unlink (track->contents.audio.filename) != 0) {
                         g_warning (_("Unable to unlink '%s'"), track->contents.audio.filename);
+                }
+
                 if (lockfile) {
                         /* remove lockfile created by mkstemp */
                         if (unlink (lockfile) != 0)
@@ -670,8 +696,9 @@ rb_recorder_finalize (GObject *object)
 
         rb_recorder_close (recorder, NULL);
 
-        if (recorder->priv->recorder)
+        if (recorder->priv->recorder != NULL) {
                 nautilus_burn_recorder_cancel (recorder->priv->recorder, FALSE);
+        }
 
         g_list_foreach (recorder->priv->tracks,
                         (GFunc)recorder_track_free,
@@ -704,6 +731,7 @@ rb_recorder_new (GError **error)
         return recorder;
 }
 
+#ifdef HAVE_GSTREAMER_0_8
 static gboolean
 tick_timeout_cb (RBRecorder *recorder)
 {
@@ -713,24 +741,13 @@ tick_timeout_cb (RBRecorder *recorder)
         double elapsed;
         double secs;
         GstFormat format = GST_FORMAT_BYTES;
-#ifdef HAVE_GSTREAMER_0_10
-	GstState  state;
-#endif
 
         g_return_val_if_fail (recorder != NULL, FALSE);
         g_return_val_if_fail (RB_IS_RECORDER (recorder), FALSE);
         g_return_val_if_fail (recorder->priv != NULL, FALSE);
         g_return_val_if_fail (recorder->priv->pipeline != NULL, FALSE);
 
-#ifdef HAVE_GSTREAMER_0_8
         if (gst_element_get_state (recorder->priv->pipeline) != GST_STATE_PLAYING) {
-#elif HAVE_GSTREAMER_0_10
-	if (!gst_element_get_state (recorder->priv->pipeline, &state,  NULL, 3 * GST_SECOND)) {
-		g_warning (_("Could not retrieve state from processing pipeline"));
-		return TRUE;
-	}
-	if (state != GST_STATE_PLAYING) {		
-#endif
                 recorder->priv->tick_timeout_id = 0;
                 if (recorder->priv->start_timer) {
                         g_timer_destroy (recorder->priv->start_timer);
@@ -739,16 +756,12 @@ tick_timeout_cb (RBRecorder *recorder)
                 return FALSE;
         }
 
-#ifdef  HAVE_GSTREAMER_0_8
         if (!gst_pad_query (recorder->priv->src_pad, GST_QUERY_POSITION, &format, &position)) {
                 g_warning (_("Could not get current track position"));
                 return TRUE;
         }
 
         if (!gst_pad_query (recorder->priv->src_pad, GST_QUERY_TOTAL, &format, &total)) {
-#elif HAVE_GSTREAMER_0_10
-		if (!gst_element_query_position (recorder->priv->src, &format, &position) || !gst_element_query_duration (recorder->priv->src, &format, &total)) {
-#endif
                 g_warning (_("Could not get current track position"));
                 return TRUE;
         }
@@ -764,10 +777,11 @@ tick_timeout_cb (RBRecorder *recorder)
 
         rate = (double)(position - recorder->priv->start_pos) / elapsed;
 
-        if (rate >= 1)
+        if (rate >= 1) {
                 secs = ceil ((total - position) / rate);
-        else
+        } else {
                 secs = -1;
+        }
 
         if (fraction != recorder->priv->progress) {
                 recorder->priv->progress = fraction;
@@ -777,13 +791,129 @@ tick_timeout_cb (RBRecorder *recorder)
                                fraction, (long)secs);
         }
 
-#ifdef HAVE_GSTREAMER_0_8
         /* Extra kick in the pants to keep things moving on a busy system */
         gst_bin_iterate (GST_BIN (recorder->priv->pipeline));
-#endif
 
         return TRUE;
 }
+#elif HAVE_GSTREAMER_0_10
+static gboolean
+tick_timeout_cb (RBRecorder *recorder)
+{
+        gint64 position, total;
+        double fraction;
+        double rate;
+        double elapsed;
+        double secs;
+        GstFormat format = GST_FORMAT_BYTES;
+	GstState  state;
+
+        g_return_val_if_fail (recorder != NULL, FALSE);
+        g_return_val_if_fail (RB_IS_RECORDER (recorder), FALSE);
+        g_return_val_if_fail (recorder->priv != NULL, FALSE);
+        g_return_val_if_fail (recorder->priv->pipeline != NULL, FALSE);
+
+	if (!gst_element_get_state (recorder->priv->pipeline, &state,  NULL, 3 * GST_SECOND)) {
+		g_warning (_("Could not retrieve state from processing pipeline"));
+		return TRUE;
+	}
+
+	if (state != GST_STATE_PLAYING) {		
+                recorder->priv->tick_timeout_id = 0;
+                if (recorder->priv->start_timer) {
+                        g_timer_destroy (recorder->priv->start_timer);
+                        recorder->priv->start_timer = NULL;
+                }
+                return FALSE;
+        }
+
+        if (!gst_element_query_position (recorder->priv->src, &format, &position) || !gst_element_query_duration (recorder->priv->src, &format, &total)) {
+                g_warning (_("Could not get current track position"));
+                return TRUE;
+        }
+
+        if (! recorder->priv->start_timer) {
+                recorder->priv->start_timer = g_timer_new ();
+                recorder->priv->start_pos = position;
+        }
+
+        fraction = (float)position / (float)total;
+
+        elapsed = g_timer_elapsed (recorder->priv->start_timer, NULL);
+
+        rate = (double)(position - recorder->priv->start_pos) / elapsed;
+
+        if (rate >= 1) {
+                secs = ceil ((total - position) / rate);
+        } else {
+                secs = -1;
+        }
+
+        if (fraction != recorder->priv->progress) {
+                recorder->priv->progress = fraction;
+                g_signal_emit (G_OBJECT (recorder),
+                               rb_recorder_signals [TRACK_PROGRESS_CHANGED],
+                               0,
+                               fraction, (long)secs);
+        }
+
+        return TRUE;
+}
+#endif
+
+
+#ifdef HAVE_GSTREAMER_0_8
+static gboolean
+rb_recorder_sync_pipeline (RBRecorder *recorder,
+                           GError    **error)
+{
+        g_return_val_if_fail (recorder != NULL, FALSE);
+        g_return_val_if_fail (RB_IS_RECORDER (recorder), FALSE);
+        g_return_val_if_fail (recorder->priv != NULL, FALSE);
+        g_return_val_if_fail (recorder->priv->pipeline != NULL, FALSE);
+        
+        rb_debug ("Syncing pipeline");
+        if (recorder->priv->playing) {
+                rb_debug ("Playing pipeline");
+                if (gst_element_set_state (recorder->priv->pipeline, GST_STATE_PLAYING) == GST_STATE_FAILURE) {
+                        g_set_error (error,
+                                     RB_RECORDER_ERROR,
+                                     RB_RECORDER_ERROR_GENERAL,
+                                     _("Could not start pipeline playing"));
+                        return FALSE;
+                }
+
+                recorder->priv->idle_id = g_idle_add ((GSourceFunc)gst_bin_iterate,
+                                                      GST_BIN (recorder->priv->pipeline));
+                recorder->priv->tick_timeout_id = g_timeout_add (200,
+                                                                 (GSourceFunc)tick_timeout_cb,
+                                                                 recorder);
+        } else {
+                rb_debug ("Pausing pipeline");
+                if (gst_element_set_state (recorder->priv->pipeline, GST_STATE_PAUSED) == GST_STATE_FAILURE) {
+                        g_set_error (error,
+                                     RB_RECORDER_ERROR,
+                                     RB_RECORDER_ERROR_GENERAL,
+                                     _("Could not pause playback"));
+                        return FALSE;
+                }
+                if (recorder->priv->idle_id > 0) {
+                        g_source_remove (recorder->priv->idle_id);
+                        recorder->priv->idle_id = 0;
+                }
+                if (recorder->priv->tick_timeout_id > 0) {
+                        g_source_remove (recorder->priv->tick_timeout_id);
+                        recorder->priv->tick_timeout_id = 0;
+                        if (recorder->priv->start_timer) {
+                                g_timer_destroy (recorder->priv->start_timer);
+                                recorder->priv->start_timer = NULL;
+                        }
+                }
+        }
+        return TRUE;
+}
+
+#elif HAVE_GSTREAMER_0_10
 
 static gboolean
 rb_recorder_sync_pipeline (RBRecorder *recorder,
@@ -797,46 +927,24 @@ rb_recorder_sync_pipeline (RBRecorder *recorder,
         rb_debug ("Syncing pipeline");
         if (recorder->priv->playing) {
                 rb_debug ("Playing pipeline");
-#ifdef HAVE_GSTREAMER_0_8
-                if (gst_element_set_state (recorder->priv->pipeline, GST_STATE_PLAYING) == GST_STATE_FAILURE) {
-#elif HAVE_GSTREAMER_0_10
                 if (gst_element_set_state (recorder->priv->pipeline, GST_STATE_PLAYING) == GST_STATE_CHANGE_FAILURE) {
-#endif
                         g_set_error (error,
                                      RB_RECORDER_ERROR,
                                      RB_RECORDER_ERROR_GENERAL,
                                      _("Could not start pipeline playing"));
                         return FALSE;
                 }
-#ifdef HAVE_GSTREAMER_0_8
-                recorder->priv->idle_id = g_idle_add ((GSourceFunc)gst_bin_iterate,
-                                                      GST_BIN (recorder->priv->pipeline));
-                recorder->priv->tick_timeout_id = g_timeout_add (200,
-                                                                 (GSourceFunc)tick_timeout_cb,
-                                                                 recorder);
-#elif HAVE_GSTREAMER_0_10
-		recorder->priv->tick_timeout_id = 
-			g_timeout_add (200, (GSourceFunc)tick_timeout_cb, recorder);
-#endif
+
+		recorder->priv->tick_timeout_id = g_timeout_add (200, (GSourceFunc)tick_timeout_cb, recorder);
         } else {
                 rb_debug ("Pausing pipeline");
-#ifdef HAVE_GSTREAMER_0_8
-                if (gst_element_set_state (recorder->priv->pipeline, GST_STATE_PAUSED) == GST_STATE_FAILURE) {
-#elif HAVE_GSTREAMER_0_10
                 if (gst_element_set_state (recorder->priv->pipeline, GST_STATE_PAUSED) == GST_STATE_CHANGE_FAILURE) {
-#endif
                         g_set_error (error,
                                      RB_RECORDER_ERROR,
                                      RB_RECORDER_ERROR_GENERAL,
                                      _("Could not pause playback"));
                         return FALSE;
                 }
-#ifdef HAVE_GSTREAMER_0_8
-                if (recorder->priv->idle_id > 0) {
-                        g_source_remove (recorder->priv->idle_id);
-                        recorder->priv->idle_id = 0;
-                }
-#endif
                 if (recorder->priv->tick_timeout_id > 0) {
                         g_source_remove (recorder->priv->tick_timeout_id);
                         recorder->priv->tick_timeout_id = 0;
@@ -848,6 +956,7 @@ rb_recorder_sync_pipeline (RBRecorder *recorder,
         }
         return TRUE;
 }
+#endif
 
 void
 rb_recorder_close (RBRecorder *recorder,
@@ -868,8 +977,9 @@ rb_recorder_close (RBRecorder *recorder,
         g_free (recorder->priv->dest_file);
         recorder->priv->dest_file = NULL;
 
-        if (recorder->priv->pipeline == NULL)
+        if (recorder->priv->pipeline == NULL) {
                 return;
+        }
 
         rb_recorder_gst_free_pipeline (recorder);
 }
@@ -932,8 +1042,9 @@ rb_recorder_open (RBRecorder   *recorder,
         }
 
         rb_recorder_construct (recorder, src_uri, error);
-        if (error && *error)
+        if (error && *error) {
                 return;
+        }
                 
 #ifdef HAVE_GSTREAMER_0_8
         if (recorder->priv->idle_id > 0) {
@@ -963,7 +1074,7 @@ rb_recorder_open (RBRecorder   *recorder,
 
         add_track (recorder, cdtext);
 
-        if (!rb_recorder_sync_pipeline (recorder, error)) {
+        if (! rb_recorder_sync_pipeline (recorder, error)) {
                 rb_recorder_close (recorder, NULL);
         }
 }
@@ -998,8 +1109,9 @@ rb_recorder_pause (RBRecorder *recorder,
         g_return_if_fail (RB_IS_RECORDER (recorder));
         g_return_if_fail (recorder->priv != NULL);
 
-        if (!recorder->priv->playing)
+        if (! recorder->priv->playing) {
                 return;
+        }
 
         recorder->priv->playing = FALSE;
 
@@ -1008,6 +1120,16 @@ rb_recorder_pause (RBRecorder *recorder,
         rb_recorder_sync_pipeline (recorder, NULL);
 }
 
+#if !NAUTILUS_BURN_CHECK_VERSION(2,15,3)
+static const char *
+nautilus_burn_drive_get_device (NautilusBurnDrive *drive)
+{
+	g_return_val_if_fail (drive != NULL, NULL);
+
+	return drive->device;
+}
+#endif
+
 char *
 rb_recorder_get_device (RBRecorder  *recorder,
                         GError     **error)
@@ -1017,8 +1139,9 @@ rb_recorder_get_device (RBRecorder  *recorder,
         g_return_val_if_fail (recorder != NULL, NULL);
         g_return_val_if_fail (RB_IS_RECORDER (recorder), NULL);
 
-        if (error)
+        if (error) {
                 *error = NULL;
+        }
 
         drive = recorder->priv->drive;
 
@@ -1030,7 +1153,7 @@ rb_recorder_get_device (RBRecorder  *recorder,
                 return NULL;
         }
 
-        return g_strdup (drive->device);
+        return g_strdup (nautilus_burn_drive_get_device (drive));
 }
 
 gboolean
@@ -1038,21 +1161,36 @@ rb_recorder_set_device (RBRecorder  *recorder,
                         const char  *device,
                         GError     **error)
 {
+#if NAUTILUS_BURN_CHECK_VERSION(2,15,3)
+        NautilusBurnDriveMonitor *monitor;
+#else
         GList *drives;
         GList *tmp;
+#endif
+        int type;
 
         g_return_val_if_fail (recorder != NULL, FALSE);
         g_return_val_if_fail (RB_IS_RECORDER (recorder), FALSE);
         g_return_val_if_fail (device != NULL, FALSE);
 
-        if (error)
+        if (error) {
                 *error = NULL;
+        }
 
-        if (recorder->priv->drive) {
+        if (recorder->priv->drive != NULL) {
                 nautilus_burn_drive_unref (recorder->priv->drive);
                 recorder->priv->drive = NULL;
         }
 
+        type = 0;
+
+#if NAUTILUS_BURN_CHECK_VERSION(2,15,3)
+        monitor = nautilus_burn_get_drive_monitor ();
+        recorder->priv->drive = nautilus_burn_drive_monitor_get_drive_for_device (monitor, device);
+        if (recorder->priv->drive != NULL) {
+                type = nautilus_burn_drive_get_drive_type (recorder->priv->drive);
+        }
+#else
         drives = nautilus_burn_drive_get_list (TRUE, FALSE);
 
         for (tmp = drives; tmp != NULL; tmp = tmp->next) {
@@ -1067,6 +1205,9 @@ rb_recorder_set_device (RBRecorder  *recorder,
         }
         g_list_free (drives);
 
+        type = recorder->priv->drive->type;
+#endif
+
         if (! recorder->priv->drive) {
                 g_set_error (error,
                              RB_RECORDER_ERROR,
@@ -1076,7 +1217,7 @@ rb_recorder_set_device (RBRecorder  *recorder,
                 return FALSE;
         }
 
-        if (! (recorder->priv->drive->type & NAUTILUS_BURN_DRIVE_TYPE_CD_RECORDER)) {
+        if (! (type & NAUTILUS_BURN_DRIVE_TYPE_CD_RECORDER)) {
                 g_set_error (error,
                              RB_RECORDER_ERROR,
                              RB_RECORDER_ERROR_GENERAL,
@@ -1224,7 +1365,7 @@ rb_recorder_get_default_device (void)
         drive = rb_recorder_get_default_drive ();
 
         if (drive) {
-                device = g_strdup (drive->device);
+                device = g_strdup (nautilus_burn_drive_get_device (drive));
                 nautilus_burn_drive_unref (drive);
         }
 
@@ -1251,7 +1392,12 @@ rb_recorder_get_media_length (RBRecorder *recorder,
                 return -1;
         }
 
+#if NAUTILUS_BURN_CHECK_VERSION(2,15,3)
+        size = nautilus_burn_drive_get_media_capacity (recorder->priv->drive);
+#else
         size = nautilus_burn_drive_get_media_size (recorder->priv->drive);
+#endif
+
         if (size > 0) {
 #ifdef NAUTILUS_BURN_DRIVE_SIZE_TO_TIME
                 secs = NAUTILUS_BURN_DRIVE_SIZE_TO_TIME (size);
@@ -1373,8 +1519,9 @@ get_tracks_length (RBRecorder *recorder,
         GList *l;
         gint64 total = 0;
 
-        if (!recorder->priv->tracks)
+        if (! recorder->priv->tracks) {
                 return -1;
+        }
 
         for (l = recorder->priv->tracks; l; l = l->next) {
                 NautilusBurnRecorderTrack *track = l->data;
@@ -1496,6 +1643,8 @@ rb_recorder_burn (RBRecorder *recorder,
                         msg = g_strdup (_("There was an error writing to the CD"));
                 }
 
+                rb_debug ("Recorder error: %s", msg);
+
                 g_set_error (error,
                              RB_RECORDER_ERROR,
                              RB_RECORDER_ERROR_GENERAL,
@@ -1510,5 +1659,7 @@ rb_recorder_burn (RBRecorder *recorder,
         g_object_unref (cdrecorder);
         recorder->priv->recorder = NULL;
 
+        rb_debug ("Recorder done: %d", result);
+
         return result;
 }
