@@ -16,7 +16,7 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
  *
  */
 
@@ -36,6 +36,12 @@
 #include "rb-marshal.h"
 #include "rb-util.h"
 
+struct ReverseSortData
+{
+	GCompareDataFunc	func;
+	gpointer		data;
+};
+
 static void rhythmdb_query_model_query_results_init (RhythmDBQueryResultsIface *iface);
 static void rhythmdb_query_model_tree_model_init (GtkTreeModelIface *iface);
 static void rhythmdb_query_model_drag_source_init (RbTreeDragSourceIface *iface);
@@ -76,8 +82,8 @@ static void rhythmdb_query_model_entry_deleted_cb (RhythmDB *db, RhythmDBEntry *
 
 static void rhythmdb_query_model_filter_out_entry (RhythmDBQueryModel *model,
 						   RhythmDBEntry *entry);
-static void rhythmdb_query_model_do_reorder (RhythmDBQueryModel *model, RhythmDBEntry *entry);
-static void rhythmdb_query_model_emit_reorder (RhythmDBQueryModel *model, gint old_pos, gint new_pos);
+static gboolean rhythmdb_query_model_do_reorder (RhythmDBQueryModel *model, RhythmDBEntry *entry);
+static gboolean rhythmdb_query_model_emit_reorder (RhythmDBQueryModel *model, gint old_pos, gint new_pos);
 static gboolean rhythmdb_query_model_drag_data_get (RbTreeDragSource *dragsource,
 							  GList *paths,
 							  GtkSelectionData *selection_data);
@@ -151,7 +157,10 @@ static void rhythmdb_query_model_base_entry_removed (RhythmDBQueryModel *base_mo
 						     RhythmDBQueryModel *model);
 static int rhythmdb_query_model_child_index_to_base_index (RhythmDBQueryModel *model, int index);
 
-static gint _reverse_sorting_func (gpointer a, gpointer b, RhythmDBQueryModel *model);
+static gint _reverse_sorting_func (gpointer a, gpointer b, struct ReverseSortData *model);
+static gboolean rhythmdb_query_model_within_limit (RhythmDBQueryModel *model,
+						   RhythmDBEntry *entry);
+static gboolean rhythmdb_query_model_reapply_query_cb (RhythmDBQueryModel *model);
 
 struct RhythmDBQueryModelUpdate
 {
@@ -187,16 +196,16 @@ struct RhythmDBQueryModelPrivate
 	RhythmDBQueryModel *base_model;
 
 	GCompareDataFunc sort_func;
-	RhythmDBPropType sort_prop_id;
+	gpointer sort_data;
+	GDestroyNotify sort_data_destroy;
 	gboolean sort_reverse;
 
 	GPtrArray *query, *original_query;
 
 	guint stamp;
 
-	guint64 max_size;
-	guint max_count;
-	guint max_time;
+	RhythmDBQueryModelLimitType limit_type;
+	GValueArray *limit_value;
 
 	glong total_duration;
 	guint64 total_size;
@@ -210,6 +219,8 @@ struct RhythmDBQueryModelPrivate
 
 	gboolean reorder_drag_and_drop;
 	gboolean show_hidden;
+
+	gint query_reapply_timeout_id;
 };
 
 #define RHYTHMDB_QUERY_MODEL_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), RHYTHMDB_TYPE_QUERY_MODEL, RhythmDBQueryModelPrivate))
@@ -220,11 +231,11 @@ enum
 	PROP_RHYTHMDB,
 	PROP_QUERY,
 	PROP_SORT_FUNC,
-	PROP_SORT_PROP_ID,
+	PROP_SORT_DATA,
+	PROP_SORT_DATA_DESTROY,
 	PROP_SORT_REVERSE,
-	PROP_MAX_SIZE,
-	PROP_MAX_COUNT,
-	PROP_MAX_TIME,
+	PROP_LIMIT_TYPE,
+	PROP_LIMIT_VALUE,
 	PROP_SHOW_HIDDEN,
 	PROP_BASE_MODEL,
 };
@@ -235,6 +246,7 @@ enum
 	ENTRY_PROP_CHANGED,
 	ENTRY_REMOVED,
 	NON_ENTRY_DROPPED,
+	POST_ENTRY_DELETE,
 	LAST_SIGNAL
 };
 
@@ -277,12 +289,17 @@ rhythmdb_query_model_class_init (RhythmDBQueryModelClass *klass)
 							      "Sort function",
 							       G_PARAM_READWRITE));
 	g_object_class_install_property (object_class,
-					 PROP_SORT_PROP_ID,
-					 g_param_spec_int ("sort-prop",
-							   "SortProp",
-							   "Sort property ID",
-							   0, RHYTHMDB_NUM_PROPERTIES, 0,
-							   G_PARAM_READWRITE));
+					 PROP_SORT_DATA,
+					 g_param_spec_pointer ("sort-data",
+							       "Sort data",
+							       "Sort data",
+							       G_PARAM_READWRITE));
+	g_object_class_install_property (object_class,
+					 PROP_SORT_DATA_DESTROY,
+					 g_param_spec_pointer ("sort-data-destroy",
+							       "Sort data destroy",
+							       "Sort data destroy function",
+							       G_PARAM_READWRITE));
 	g_object_class_install_property (object_class,
 					 PROP_SORT_REVERSE,
 					 g_param_spec_boolean ("sort-reverse",
@@ -291,26 +308,20 @@ rhythmdb_query_model_class_init (RhythmDBQueryModelClass *klass)
 							      FALSE,
 							      G_PARAM_READWRITE));
 	g_object_class_install_property (object_class,
-					 PROP_MAX_SIZE,
-					 g_param_spec_int ("max-size",
-							   "maxsize",
-							   "maximum size (MB)",
-							   0, G_MAXINT, 0,
-							   G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+					 PROP_LIMIT_TYPE,
+					 g_param_spec_enum ("limit-type",
+							    "limit-type",
+							    "type of limit",
+							    RHYTHMDB_TYPE_QUERY_MODEL_LIMIT_TYPE,
+							    RHYTHMDB_QUERY_MODEL_LIMIT_NONE,
+							    G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 	g_object_class_install_property (object_class,
-					 PROP_MAX_COUNT,
-					 g_param_spec_int ("max-count",
-							   "maxcount",
-							   "maximum count (songs)",
-							   0, G_MAXINT, 0,
-							   G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
-	g_object_class_install_property (object_class,
-					 PROP_MAX_TIME,
-					 g_param_spec_int ("max-time",
-							   "maxtime",
-							   "maximum time (seconds)",
-							   0, G_MAXINT, 0,
-							   G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+					 PROP_LIMIT_VALUE,
+					 g_param_spec_boxed ("limit-value",
+							     "limit-value",
+							     "value of limit",
+							     G_TYPE_VALUE_ARRAY,
+							     G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
 	g_object_class_install_property (object_class,
 					 PROP_SHOW_HIDDEN,
 					 g_param_spec_boolean ("show-hidden",
@@ -361,6 +372,15 @@ rhythmdb_query_model_class_init (RhythmDBQueryModelClass *klass)
 			      NULL, NULL,
 			      g_cclosure_marshal_VOID__VOID,
 			      G_TYPE_NONE, 0);
+	rhythmdb_query_model_signals[POST_ENTRY_DELETE] =
+		g_signal_new ("post-entry-delete",
+			      RHYTHMDB_TYPE_QUERY_MODEL,
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (RhythmDBQueryModelClass, post_entry_delete),
+			      NULL, NULL,
+			      g_cclosure_marshal_VOID__POINTER,
+			      G_TYPE_NONE,
+			      1, RHYTHMDB_TYPE_ENTRY);
 
 	g_type_class_add_private (klass, sizeof (RhythmDBQueryModelPrivate));
 }
@@ -423,102 +443,50 @@ rhythmdb_query_model_set_property (GObject *object,
 		model->priv->query = rhythmdb_query_copy (g_value_get_pointer (value));
 		model->priv->original_query = rhythmdb_query_copy (model->priv->query);
 		rhythmdb_query_preprocess (model->priv->db, model->priv->query);
+
+		/* if the query contains time-relative criteria, re-run it periodically.
+		 * currently it's just every half hour, but perhaps it could be smarter.
+		 */
+		if (rhythmdb_query_is_time_relative (model->priv->db, model->priv->query)) {
+			if (model->priv->query_reapply_timeout_id == 0) {
+				model->priv->query_reapply_timeout_id = 
+					g_timeout_add (1 * 60 * 1000,
+						       (GSourceFunc) rhythmdb_query_model_reapply_query_cb,
+						       model);
+			}
+		} else if (model->priv->query_reapply_timeout_id) {
+			g_source_remove (model->priv->query_reapply_timeout_id);
+			model->priv->query_reapply_timeout_id = 0;
+		}
 		break;
 	case PROP_SORT_FUNC:
 		model->priv->sort_func = g_value_get_pointer (value);
 		break;
-	case PROP_SORT_PROP_ID:
-		model->priv->sort_prop_id = g_value_get_int (value);
+	case PROP_SORT_DATA:
+		if (model->priv->sort_data_destroy && model->priv->sort_data)
+			model->priv->sort_data_destroy (model->priv->sort_data);
+		model->priv->sort_data = g_value_get_pointer (value);
+		break;
+	case PROP_SORT_DATA_DESTROY:
+		model->priv->sort_data_destroy = g_value_get_pointer (value);
 		break;
 	case PROP_SORT_REVERSE:
 		model->priv->sort_reverse  = g_value_get_boolean (value);
 		break;
-	case PROP_MAX_SIZE:
-		model->priv->max_size = g_value_get_int (value) * 1024 * 1024;
+	case PROP_LIMIT_TYPE:
+		model->priv->limit_type = g_value_get_enum (value);
 		break;
-	case PROP_MAX_COUNT:
-		model->priv->max_count = g_value_get_int (value);
-		break;
-	case PROP_MAX_TIME:
-		model->priv->max_time = g_value_get_int (value);
+	case PROP_LIMIT_VALUE:
+		if (model->priv->limit_value)
+			g_value_array_free (model->priv->limit_value);
+		model->priv->limit_value = (GValueArray*)g_value_dup_boxed (value);
 		break;
 	case PROP_SHOW_HIDDEN:
 		model->priv->show_hidden = g_value_get_boolean (value);
 		/* FIXME: this will have funky issues if this is changed after entries are added */
 		break;
 	case PROP_BASE_MODEL:
-		if (model->priv->base_model) {
-			g_signal_handlers_disconnect_by_func (G_OBJECT (model->priv->base_model),
-							      G_CALLBACK (rhythmdb_query_model_base_row_inserted),
-							      model);
-			g_signal_handlers_disconnect_by_func (G_OBJECT (model->priv->base_model),
-							      G_CALLBACK (rhythmdb_query_model_base_row_deleted),
-							      model);
-			g_signal_handlers_disconnect_by_func (G_OBJECT (model->priv->base_model),
-							      G_CALLBACK (rhythmdb_query_model_base_non_entry_dropped),
-							      model);
-			g_signal_handlers_disconnect_by_func (G_OBJECT (model->priv->base_model),
-							      G_CALLBACK (rhythmdb_query_model_base_complete),
-							      model);
-			g_signal_handlers_disconnect_by_func (G_OBJECT (model->priv->base_model),
-							      G_CALLBACK (rhythmdb_query_model_base_rows_reordered),
-							      model);
-			g_signal_handlers_disconnect_by_func (G_OBJECT (model->priv->base_model),
-							      G_CALLBACK (rhythmdb_query_model_base_entry_removed),
-							      model);
-			g_object_unref (model->priv->base_model);
-		}
-
-		model->priv->base_model = g_value_get_object (value);
-
-		if (model->priv->base_model) {
-			g_object_ref (model->priv->base_model);
-			g_assert (model->priv->base_model->priv->db == model->priv->db);
-
-			g_signal_connect_object (G_OBJECT (model->priv->base_model),
-						 "row-inserted",
-						 G_CALLBACK (rhythmdb_query_model_base_row_inserted),
-						 model, 0);
-			g_signal_connect_object (G_OBJECT (model->priv->base_model),
-						 "row-deleted",
-						 G_CALLBACK (rhythmdb_query_model_base_row_deleted),
-						 model, 0);
-			g_signal_connect_object (G_OBJECT (model->priv->base_model),
-						 "non-entry-dropped",
-						 G_CALLBACK (rhythmdb_query_model_base_non_entry_dropped),
-						 model, 0);
-			g_signal_connect_object (G_OBJECT (model->priv->base_model),
-						 "complete",
-						 G_CALLBACK (rhythmdb_query_model_base_complete),
-						 model, 0);
-			g_signal_connect_object (G_OBJECT (model->priv->base_model),
-						 "rows-reordered",
-						 G_CALLBACK (rhythmdb_query_model_base_rows_reordered),
-						 model, 0);
-			g_signal_connect_object (G_OBJECT (model->priv->base_model),
-						 "entry-removed",
-						 G_CALLBACK (rhythmdb_query_model_base_entry_removed),
-						 model, 0);
-
-			if (model->priv->base_model->priv->entries) {
-				GSequencePtr ptr, next;
-				RhythmDBEntry *entry;
-
-				ptr = g_sequence_get_begin_ptr (model->priv->base_model->priv->entries);
-				while (!g_sequence_ptr_is_end (ptr)) {
-					next = g_sequence_ptr_next (ptr);
-					entry = (RhythmDBEntry *)g_sequence_ptr_get_data (ptr);
-					if (model->priv->query == NULL ||
-					    rhythmdb_evaluate_query (model->priv->db, model->priv->query, entry)) {
-						if (model->priv->show_hidden || (rhythmdb_entry_get_boolean (entry, RHYTHMDB_PROP_HIDDEN) == FALSE))
-							rhythmdb_query_model_do_insert (model, entry, -1);
-					}
-
-					ptr = next;
-				}
-			}
-		}
-		
+		rhythmdb_query_model_chain (model, g_value_get_object (value), TRUE);
 		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
@@ -545,20 +513,20 @@ rhythmdb_query_model_get_property (GObject *object,
 	case PROP_SORT_FUNC:
 		g_value_set_pointer (value, model->priv->sort_func);
 		break;
-	case PROP_SORT_PROP_ID:
-		g_value_set_int (value, model->priv->sort_prop_id);
+	case PROP_SORT_DATA:
+		g_value_set_pointer (value, model->priv->sort_data);
+		break;
+	case PROP_SORT_DATA_DESTROY:
+		g_value_set_pointer (value, model->priv->sort_data_destroy);
 		break;
 	case PROP_SORT_REVERSE:
 		g_value_set_boolean (value, model->priv->sort_reverse);
 		break;
-	case PROP_MAX_SIZE:
-		g_value_set_int (value, model->priv->max_size / (1024 * 1024));
+	case PROP_LIMIT_TYPE:
+		g_value_set_enum (value, model->priv->limit_type);
 		break;
-	case PROP_MAX_COUNT:
-		g_value_set_int (value, model->priv->max_count);
-		break;
-	case PROP_MAX_TIME:
-		g_value_set_int (value, model->priv->max_time);
+	case PROP_LIMIT_VALUE:
+		g_value_set_boxed (value, model->priv->limit_value);
 		break;
 	case PROP_SHOW_HIDDEN:
 		g_value_set_boolean (value, model->priv->show_hidden);
@@ -684,19 +652,30 @@ rhythmdb_query_model_finalize (GObject *object)
 	if (model->priv->original_query)
 		rhythmdb_query_free (model->priv->original_query);
 
+	if (model->priv->sort_data_destroy && model->priv->sort_data)
+		model->priv->sort_data_destroy (model->priv->sort_data);
+
+	if (model->priv->limit_value)
+		g_value_array_free (model->priv->limit_value);
+
+	if (model->priv->query_reapply_timeout_id)
+		g_source_remove (model->priv->query_reapply_timeout_id);
+
 	G_OBJECT_CLASS (rhythmdb_query_model_parent_class)->finalize (object);
 }
 
 RhythmDBQueryModel *
 rhythmdb_query_model_new (RhythmDB *db, GPtrArray *query,
 			  GCompareDataFunc sort_func,
-			  RhythmDBPropType sort_prop_id,
+			  gpointer sort_data,
+			  GDestroyNotify sort_data_destroy,
 			  gboolean sort_reverse)
 {
 	RhythmDBQueryModel *model = g_object_new (RHYTHMDB_TYPE_QUERY_MODEL,
 						  "db", db, "query", query,
 						  "sort-func", sort_func,
-						  "sort-prop", sort_prop_id,
+						  "sort-data", sort_data,
+						  "sort-data-destroy", sort_data_destroy,
 						  "sort-reverse", sort_reverse,
 						  NULL);
 
@@ -712,6 +691,86 @@ rhythmdb_query_model_new_empty (RhythmDB *db)
 			     "db", db, NULL);
 }
 
+void
+rhythmdb_query_model_chain (RhythmDBQueryModel *model,
+			    RhythmDBQueryModel *base,
+			    gboolean import_entries)
+{
+	rb_debug ("query model %p chaining to base model %p", model, base);
+	
+	if (model->priv->base_model) {
+		g_signal_handlers_disconnect_by_func (G_OBJECT (model->priv->base_model),
+						      G_CALLBACK (rhythmdb_query_model_base_row_inserted),
+						      model);
+		g_signal_handlers_disconnect_by_func (G_OBJECT (model->priv->base_model),
+						      G_CALLBACK (rhythmdb_query_model_base_row_deleted),
+						      model);
+		g_signal_handlers_disconnect_by_func (G_OBJECT (model->priv->base_model),
+						      G_CALLBACK (rhythmdb_query_model_base_non_entry_dropped),
+						      model);
+		g_signal_handlers_disconnect_by_func (G_OBJECT (model->priv->base_model),
+						      G_CALLBACK (rhythmdb_query_model_base_complete),
+						      model);
+		g_signal_handlers_disconnect_by_func (G_OBJECT (model->priv->base_model),
+						      G_CALLBACK (rhythmdb_query_model_base_rows_reordered),
+						      model);
+		g_signal_handlers_disconnect_by_func (G_OBJECT (model->priv->base_model),
+						      G_CALLBACK (rhythmdb_query_model_base_entry_removed),
+						      model);
+		g_object_unref (model->priv->base_model);
+	}
+
+	model->priv->base_model = base;
+
+	if (model->priv->base_model) {
+		g_object_ref (model->priv->base_model);
+		g_assert (model->priv->base_model->priv->db == model->priv->db);
+
+		g_signal_connect_object (G_OBJECT (model->priv->base_model),
+					 "row-inserted",
+					 G_CALLBACK (rhythmdb_query_model_base_row_inserted),
+					 model, 0);
+		g_signal_connect_object (G_OBJECT (model->priv->base_model),
+					 "row-deleted",
+					 G_CALLBACK (rhythmdb_query_model_base_row_deleted),
+					 model, 0);
+		g_signal_connect_object (G_OBJECT (model->priv->base_model),
+					 "non-entry-dropped",
+					 G_CALLBACK (rhythmdb_query_model_base_non_entry_dropped),
+					 model, 0);
+		g_signal_connect_object (G_OBJECT (model->priv->base_model),
+					 "complete",
+					 G_CALLBACK (rhythmdb_query_model_base_complete),
+					 model, 0);
+		g_signal_connect_object (G_OBJECT (model->priv->base_model),
+					 "rows-reordered",
+					 G_CALLBACK (rhythmdb_query_model_base_rows_reordered),
+					 model, 0);
+		g_signal_connect_object (G_OBJECT (model->priv->base_model),
+					 "entry-removed",
+					 G_CALLBACK (rhythmdb_query_model_base_entry_removed),
+					 model, 0);
+
+		if (import_entries && model->priv->base_model->priv->entries) {
+			GSequencePtr ptr, next;
+			RhythmDBEntry *entry;
+
+			ptr = g_sequence_get_begin_ptr (model->priv->base_model->priv->entries);
+			while (!g_sequence_ptr_is_end (ptr)) {
+				next = g_sequence_ptr_next (ptr);
+				entry = (RhythmDBEntry *)g_sequence_ptr_get_data (ptr);
+				if (model->priv->query == NULL ||
+				    rhythmdb_evaluate_query (model->priv->db, model->priv->query, entry)) {
+					if (model->priv->show_hidden || (rhythmdb_entry_get_boolean (entry, RHYTHMDB_PROP_HIDDEN) == FALSE))
+						rhythmdb_query_model_do_insert (model, entry, -1);
+				}
+
+				ptr = next;
+			}
+		}
+	}
+}
+
 gboolean
 rhythmdb_query_model_has_pending_changes (RhythmDBQueryModel *model)
 {
@@ -807,7 +866,19 @@ rhythmdb_query_model_entry_changed_cb (RhythmDB *db, RhythmDBEntry *entry,
 	}
 
 	/* it may have moved, so we can't just emit a changed entry */
-	rhythmdb_query_model_do_reorder (model, entry);
+	if (!rhythmdb_query_model_do_reorder (model, entry)) {
+		/* but if it didn't, we can */
+		GtkTreeIter iter;
+		GtkTreePath *path;
+
+		if (rhythmdb_query_model_entry_to_iter (model, entry, &iter)) {
+			path = rhythmdb_query_model_get_path (GTK_TREE_MODEL (model),
+							      &iter);
+			gtk_tree_model_row_changed (GTK_TREE_MODEL (model),
+						     path, &iter);
+			gtk_tree_path_free (path);
+		}
+	}
 }
 
 static void
@@ -919,17 +990,26 @@ static void
 rhythmdb_query_model_insert_into_main_list (RhythmDBQueryModel *model, RhythmDBEntry *entry, gint index)
 {
 	GSequencePtr ptr;
-	GCompareDataFunc sort_func = model->priv->sort_func;
-
 	rhythmdb_entry_ref (model->priv->db, entry);
 
-	if (sort_func) {
-		if (model->priv->sort_reverse)
+	if (model->priv->sort_func) {
+		GCompareDataFunc sort_func;
+		gpointer sort_data;
+		struct ReverseSortData reverse_data;
+
+		if (model->priv->sort_reverse) {
 			sort_func = (GCompareDataFunc) _reverse_sorting_func;
+			sort_data = &reverse_data;
+			reverse_data.func = model->priv->sort_func;
+			reverse_data.data = model->priv->sort_data;
+		} else {
+			sort_func = model->priv->sort_func;
+			sort_data = model->priv->sort_data;
+		}
 
 		ptr = g_sequence_insert_sorted (model->priv->entries, entry,
 						sort_func,
-						model);
+						sort_data);
 	} else {
 		if (index == -1) {
 			ptr = g_sequence_get_end_ptr (model->priv->entries);
@@ -951,17 +1031,27 @@ static void
 rhythmdb_query_model_insert_into_limited_list (RhythmDBQueryModel *model, RhythmDBEntry *entry)
 {
 	GSequencePtr ptr;
-	GCompareDataFunc sort_func = model->priv->sort_func;
 
 	rhythmdb_entry_ref (model->priv->db, entry);
 
-	if (sort_func) {
-		if (model->priv->sort_reverse)
+	if (model->priv->sort_func) {
+		GCompareDataFunc sort_func;
+		gpointer sort_data;
+		struct ReverseSortData reverse_data;
+
+		if (model->priv->sort_reverse) {
 			sort_func = (GCompareDataFunc) _reverse_sorting_func;
+			sort_data = &reverse_data;
+			reverse_data.func = model->priv->sort_func;
+			reverse_data.data = model->priv->sort_data;
+		} else {
+			sort_func = model->priv->sort_func;
+			sort_data = model->priv->sort_data;
+		}
 
 		ptr = g_sequence_insert_sorted (model->priv->limited_entries, entry,
 						sort_func,
-						model);
+						sort_data);
 	} else {
 		ptr = g_sequence_get_end_ptr (model->priv->limited_entries);
 		g_sequence_insert (ptr, entry);
@@ -991,6 +1081,8 @@ rhythmdb_query_model_remove_from_main_list (RhythmDBQueryModel *model, RhythmDBE
 
 	g_sequence_remove (ptr);
 	g_assert (g_hash_table_remove (model->priv->reverse_map, entry));
+
+	g_signal_emit (G_OBJECT (model), rhythmdb_query_model_signals[POST_ENTRY_DELETE], 0, entry);
 	
 	rhythmdb_entry_unref (model->priv->db, entry);
 }
@@ -1012,10 +1104,7 @@ rhythmdb_query_model_update_limited_entries (RhythmDBQueryModel *model)
 	GSequencePtr ptr;
 
 	/* make it fit inside the limits */
-	while ((model->priv->max_count > 0 && g_hash_table_size (model->priv->reverse_map) > model->priv->max_count) ||
-	      (model->priv->max_size > 0 && model->priv->total_size > model->priv->max_size) ||
-	      (model->priv->max_time > 0 && model->priv->total_duration > model->priv->max_time)) {
-
+	while (!rhythmdb_query_model_within_limit (model, NULL)) {
 		ptr = g_sequence_ptr_prev (g_sequence_get_end_ptr (model->priv->entries));
 		entry = (RhythmDBEntry*) g_sequence_ptr_get_data (ptr);
 
@@ -1025,8 +1114,6 @@ rhythmdb_query_model_update_limited_entries (RhythmDBQueryModel *model)
 
 	/* move entries that were previously limited, back to the main list */
 	while (TRUE) {
-		int size;
-		int duration;
 		GtkTreePath *path;
 		GtkTreeIter iter;
 
@@ -1037,12 +1124,7 @@ rhythmdb_query_model_update_limited_entries (RhythmDBQueryModel *model)
 		if (!entry)
 			break;
 
-		size = rhythmdb_entry_get_uint64 (entry, RHYTHMDB_PROP_FILE_SIZE);
-		duration = rhythmdb_entry_get_ulong (entry, RHYTHMDB_PROP_DURATION);
-
-		if ((model->priv->max_count > 0 && (g_hash_table_size (model->priv->reverse_map) + 1) > model->priv->max_count) ||
-		    (model->priv->max_size > 0 && model->priv->total_size + size > model->priv->max_size) ||
-		    (model->priv->max_time > 0 && model->priv->total_duration + duration > model->priv->max_time))
+		if (!rhythmdb_query_model_within_limit (model, entry))
 			break;
 
 		rhythmdb_query_model_remove_from_limited_list (model, entry);
@@ -1059,7 +1141,7 @@ rhythmdb_query_model_update_limited_entries (RhythmDBQueryModel *model)
 	}
 }
 
-static void
+static gboolean
 rhythmdb_query_model_emit_reorder (RhythmDBQueryModel *model, gint old_pos, gint new_pos)
 {
 	int length, i;
@@ -1069,7 +1151,7 @@ rhythmdb_query_model_emit_reorder (RhythmDBQueryModel *model, gint old_pos, gint
 
 	if (new_pos == old_pos) {
 		/* it hasn't moved, don't emit a re-order */
-		return;
+		return FALSE;
 	}
 
 	length = g_sequence_get_length (model->priv->entries);
@@ -1103,33 +1185,44 @@ rhythmdb_query_model_emit_reorder (RhythmDBQueryModel *model, gint old_pos, gint
 				       reorder_map);
 	gtk_tree_path_free (path);
 	free (reorder_map);
+	return TRUE;
 }
 
-static void
+static gboolean
 rhythmdb_query_model_do_reorder (RhythmDBQueryModel *model, RhythmDBEntry *entry)
 {
 	GSequencePtr ptr;
 	int old_pos, new_pos;
 	GtkTreePath *path;
 	GtkTreeIter iter;
-	GCompareDataFunc sort_func = model->priv->sort_func;
+	GCompareDataFunc sort_func;
+	gpointer sort_data;
+	struct ReverseSortData reverse_data;
 
-	if (sort_func == NULL)
-		return;
-	if (model->priv->sort_reverse)
+	if (model->priv->sort_func == NULL)
+		return FALSE;
+
+	if (model->priv->sort_reverse) {
 		sort_func = (GCompareDataFunc) _reverse_sorting_func;
+		sort_data = &reverse_data;
+		reverse_data.func = model->priv->sort_func;
+		reverse_data.data = model->priv->sort_data;
+	} else {
+		sort_func = model->priv->sort_func;
+		sort_data = model->priv->sort_data;
+	}
 
 	ptr = g_sequence_get_begin_ptr (model->priv->limited_entries);
 
 	if (ptr != NULL && !g_sequence_ptr_is_end (ptr)) {
 		RhythmDBEntry *first_limited = g_sequence_ptr_get_data (ptr);
-		int cmp = (sort_func) (entry, first_limited, model);
+		int cmp = (sort_func) (entry, first_limited, sort_data);
 
 		if (cmp > 0) {
 			/* the entry belongs in the limited list, so we don't need a re-order */
 			rhythmdb_query_model_remove_entry (model, entry);
 			rhythmdb_query_model_do_insert (model, entry, -1);
-			return;
+			return TRUE;
 		}
 	}
 
@@ -1149,11 +1242,11 @@ rhythmdb_query_model_do_reorder (RhythmDBQueryModel *model, RhythmDBEntry *entry
 
 	ptr = g_sequence_insert_sorted (model->priv->entries, entry,
 					sort_func,
-					model);
+					sort_data);
 	new_pos = g_sequence_ptr_get_position (ptr);
 	g_hash_table_insert (model->priv->reverse_map, entry, ptr);
 
-	rhythmdb_query_model_emit_reorder (model, old_pos, new_pos);
+	return rhythmdb_query_model_emit_reorder (model, old_pos, new_pos);
 }
 
 static void
@@ -1398,6 +1491,7 @@ rhythmdb_query_model_drag_data_get (RbTreeDragSource *dragsource,
 		RhythmDBEntry *entry;
 		GList *tem;
 		GString *data;
+		gboolean need_newline = FALSE;
 
 		data = g_string_new ("");
 
@@ -1413,10 +1507,14 @@ rhythmdb_query_model_drag_data_get (RbTreeDragSource *dragsource,
 			entry = g_sequence_ptr_get_data (iter.user_data);
 
 			location = rhythmdb_entry_get_string (entry, RHYTHMDB_PROP_LOCATION);
-			g_string_append (data, location);
+			if (location == NULL)
+				continue;
 
-			if (tem->next)
+			if (need_newline)
 				g_string_append (data, "\r\n");
+
+			g_string_append (data, location);
+			need_newline = TRUE;
 		}
 
 		gtk_selection_data_set (selection_data,
@@ -1802,11 +1900,15 @@ rhythmdb_query_model_iter_parent (GtkTreeModel *tree_model,
 }
 
 char *
-rhythmdb_query_model_compute_status_normal (RhythmDBQueryModel *model)
+rhythmdb_query_model_compute_status_normal (RhythmDBQueryModel *model, 
+					    const char *singular, 
+					    const char *plural)
 {
 	return rhythmdb_compute_status_normal (gtk_tree_model_iter_n_children (GTK_TREE_MODEL (model), NULL),
 					       rhythmdb_query_model_get_duration (model),
-					       rhythmdb_query_model_get_size (model));
+					       rhythmdb_query_model_get_size (model),
+					       singular,
+					       plural);
 }
 
 static void
@@ -1828,11 +1930,13 @@ apply_updated_entry_sequence (RhythmDBQueryModel *model, GSequence *new_entries)
 		GSequencePtr old_ptr;
 
 		old_ptr = g_hash_table_lookup (model->priv->reverse_map, entry);
-		reorder_map[i] = g_sequence_ptr_get_position (ptr);
+		reorder_map[i] = g_sequence_ptr_get_position (old_ptr);
 		g_hash_table_replace (model->priv->reverse_map, entry, ptr);
 
 		ptr = g_sequence_ptr_next (ptr);
 	}
+	g_sequence_free (model->priv->entries);
+	model->priv->entries = new_entries;
 
 	/* emit the re-order and clean up */
 	gtk_tree_model_get_iter_first (GTK_TREE_MODEL (model), &iter);
@@ -1843,38 +1947,45 @@ apply_updated_entry_sequence (RhythmDBQueryModel *model, GSequence *new_entries)
 
 	gtk_tree_path_free (path);
 	free (reorder_map);
-	g_sequence_free (model->priv->entries);
-	model->priv->entries = new_entries;
 }
 
 void
 rhythmdb_query_model_set_sort_order (RhythmDBQueryModel *model,
 				     GCompareDataFunc sort_func,
-				     RhythmDBPropType sort_prop_id,
+				     gpointer sort_data,
+				     GDestroyNotify sort_data_destroy,
 				     gboolean sort_reverse)
 {
 	GSequence *new_entries;
 	GSequencePtr ptr;
 	int length, i;
+	struct ReverseSortData reverse_data;
 
 	if ((model->priv->sort_func == sort_func) &&
-	    (model->priv->sort_prop_id == sort_prop_id) &&
+	    (model->priv->sort_data == sort_data) &&
+	    (model->priv->sort_data_destroy == sort_data_destroy) &&
 	    (model->priv->sort_reverse == sort_reverse))
 		return;
 	
-	g_return_if_fail (((model->priv->max_count == 0) &&
-			  (model->priv->max_time == 0) &&
-			  (model->priv->max_size == 0)) ||
+	g_return_if_fail ((model->priv->limit_type == RHYTHMDB_QUERY_MODEL_LIMIT_NONE) ||
 			  (model->priv->sort_func == NULL));
 	if (model->priv->sort_func == NULL)
 		g_assert (g_sequence_get_length (model->priv->limited_entries) == 0);
 	
+	if (model->priv->sort_data_destroy && model->priv->sort_data)
+		model->priv->sort_data_destroy (model->priv->sort_data);
+	
 	model->priv->sort_func = sort_func;
-	model->priv->sort_prop_id = sort_prop_id;
+	model->priv->sort_data = sort_data;
+	model->priv->sort_data_destroy = sort_data_destroy;
 	model->priv->sort_reverse = sort_reverse;
 
-	if (sort_reverse)
+	if (model->priv->sort_reverse) {
+		reverse_data.func = sort_func;
+		reverse_data.data = sort_data;
 		sort_func = (GCompareDataFunc) _reverse_sorting_func;
+		sort_data = &reverse_data;
+	}
 
 	/* create the new sorted entry sequence */
 	length = g_sequence_get_length (model->priv->entries);
@@ -1886,7 +1997,7 @@ rhythmdb_query_model_set_sort_order (RhythmDBQueryModel *model,
 
 			g_sequence_insert_sorted (new_entries, entry,
 						  sort_func,
-						  model);
+						  sort_data);
 			ptr = g_sequence_ptr_next (ptr);
 		}
 
@@ -2131,14 +2242,14 @@ rhythmdb_query_model_reapply_query (RhythmDBQueryModel *model, gboolean filter)
 
 static gint
 _reverse_sorting_func (gpointer a, gpointer b, 
-		       RhythmDBQueryModel *model)
+		       struct ReverseSortData *reverse_data)
 {
-	return - model->priv->sort_func (a, b, model);
+	return - reverse_data->func (a, b, reverse_data->data);
 }
 
 gint
 rhythmdb_query_model_location_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
-					 RhythmDBQueryModel *model)
+					 gpointer data)
 {
 	const char *a_val;
 	const char *b_val;
@@ -2159,7 +2270,7 @@ rhythmdb_query_model_location_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
 
 gint
 rhythmdb_query_model_title_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
-				      RhythmDBQueryModel *model)
+				      gpointer data)
 {
 	const char *a_val;
 	const char *b_val;
@@ -2182,17 +2293,17 @@ rhythmdb_query_model_title_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
 	if (ret != 0)
 		return ret;
 	else
-		return rhythmdb_query_model_location_sort_func (a, b, model);
+		return rhythmdb_query_model_location_sort_func (a, b, data);
 }
 
 gint
 rhythmdb_query_model_album_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
-				      RhythmDBQueryModel *model)
+				      gpointer data)
 {
 	const char *a_val;
 	const char *b_val;
-	ulong a_num;
-	ulong b_num;
+	gulong a_num;
+	gulong b_num;
 	gint ret;
 
 	/* Sort by album name */
@@ -2236,13 +2347,13 @@ rhythmdb_query_model_album_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
 	} else if (b_val == NULL)
 		return 1;
 	else
-		return rhythmdb_query_model_location_sort_func (a, b, model);
+		return rhythmdb_query_model_location_sort_func (a, b, data);
 }
 
 
 gint
 rhythmdb_query_model_artist_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
-				       RhythmDBQueryModel *model)
+				       gpointer data)
 {
 	const char *a_val;
 	const char *b_val;
@@ -2264,12 +2375,12 @@ rhythmdb_query_model_artist_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
 	if (ret != 0)
 		return ret;
 	else
-		return rhythmdb_query_model_album_sort_func (a, b, model);
+		return rhythmdb_query_model_album_sort_func (a, b, data);
 }
 
 gint
 rhythmdb_query_model_genre_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
-				      RhythmDBQueryModel *model)
+				      gpointer data)
 {
 	const char *a_val;
 	const char *b_val;
@@ -2291,51 +2402,56 @@ rhythmdb_query_model_genre_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
 	if (ret != 0)
 		return ret;
 	else
-		return rhythmdb_query_model_artist_sort_func (a, b, model);
+		return rhythmdb_query_model_artist_sort_func (a, b, data);
 }
 
 
 gint
 rhythmdb_query_model_track_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
-				      RhythmDBQueryModel *model)
+				      gpointer data)
 {
-	return rhythmdb_query_model_album_sort_func (a, b, model);
+	return rhythmdb_query_model_album_sort_func (a, b, data);
 }
 
 
 gint
 rhythmdb_query_model_double_ceiling_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
-					       RhythmDBQueryModel *model)
+					       gpointer data)
 {
 	gdouble a_val, b_val;
+	RhythmDBPropType prop_id;
 
-	a_val = ceil (rhythmdb_entry_get_double (a, model->priv->sort_prop_id));
-	b_val = ceil (rhythmdb_entry_get_double (b, model->priv->sort_prop_id));
+	prop_id = (RhythmDBPropType) GPOINTER_TO_INT (data);
+
+	a_val = ceil (rhythmdb_entry_get_double (a, prop_id));
+	b_val = ceil (rhythmdb_entry_get_double (b, prop_id));
 
 	if (a_val != b_val)
 		return (a_val > b_val ? 1 : -1);
 	else
-		return rhythmdb_query_model_location_sort_func (a, b, model);
+		return rhythmdb_query_model_location_sort_func (a, b, data);
 }
 
 gint
 rhythmdb_query_model_ulong_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
-				      RhythmDBQueryModel *model)
+				      gpointer data)
 {
 	gulong a_val, b_val;
+	RhythmDBPropType prop_id;
 
-	a_val = rhythmdb_entry_get_ulong (a, model->priv->sort_prop_id);
-	b_val = rhythmdb_entry_get_ulong (b, model->priv->sort_prop_id);
+	prop_id = (RhythmDBPropType) GPOINTER_TO_INT (data);
+	a_val = rhythmdb_entry_get_ulong (a, prop_id);
+	b_val = rhythmdb_entry_get_ulong (b, prop_id);
 
 	if (a_val != b_val)
 		return (a_val > b_val ? 1 : -1);
 	else
-		return rhythmdb_query_model_location_sort_func (a, b, model);
+		return rhythmdb_query_model_location_sort_func (a, b, data);
 }
 
 gint
 rhythmdb_query_model_date_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
-				     RhythmDBQueryModel *model)
+				     gpointer data)
 			      
 {
 	gulong a_val, b_val;
@@ -2350,19 +2466,21 @@ rhythmdb_query_model_date_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
 	else if (a_val < b_val)
 		return -1;
 	else
-		return rhythmdb_query_model_album_sort_func (a, b, model);
+		return rhythmdb_query_model_album_sort_func (a, b, data);
 }
 
 gint
 rhythmdb_query_model_string_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
-				       RhythmDBQueryModel *model)
+				       gpointer data)
 {
 	const char *a_val;
 	const char *b_val;
 	gint ret;
+	RhythmDBPropType prop_id;
 
-	a_val = rhythmdb_entry_get_string (a, model->priv->sort_prop_id);
-	b_val = rhythmdb_entry_get_string (b, model->priv->sort_prop_id);
+	prop_id = (RhythmDBPropType) GPOINTER_TO_INT (data);
+	a_val = rhythmdb_entry_get_string (a, prop_id);
+	b_val = rhythmdb_entry_get_string (b, prop_id);
 
 	if (a_val == NULL) {
 		if (b_val == NULL)
@@ -2377,8 +2495,101 @@ rhythmdb_query_model_string_sort_func (RhythmDBEntry *a, RhythmDBEntry *b,
 	if (ret != 0)
 		return ret;
 	else
-		return rhythmdb_query_model_location_sort_func (a, b, model);
+		return rhythmdb_query_model_location_sort_func (a, b, data);
 }
 
+static gboolean
+rhythmdb_query_model_within_limit (RhythmDBQueryModel *model, RhythmDBEntry *entry)
+{
+	gboolean result = TRUE;
 
+	switch (model->priv->limit_type) {
+	case RHYTHMDB_QUERY_MODEL_LIMIT_NONE:
+		result = TRUE;
+		break;
 
+	case RHYTHMDB_QUERY_MODEL_LIMIT_COUNT:
+		{
+			gulong limit_count;
+			gulong current_count;
+			
+			limit_count = g_value_get_ulong (g_value_array_get_nth (model->priv->limit_value, 0));
+			current_count = g_hash_table_size (model->priv->reverse_map);
+
+			if (entry)
+				current_count++;
+
+			result = (current_count <= limit_count);
+			break;
+		}
+
+	case RHYTHMDB_QUERY_MODEL_LIMIT_SIZE:
+		{
+			guint64 limit_size;
+			guint64 current_size;
+			
+			limit_size = g_value_get_uint64 (g_value_array_get_nth (model->priv->limit_value, 0));
+			current_size = model->priv->total_size;
+
+			if (entry)
+				current_size += rhythmdb_entry_get_uint64 (entry, RHYTHMDB_PROP_FILE_SIZE);
+
+			/* the limit is in MB */
+			result = (current_size / (1024 * 1024) <= limit_size);
+			break;
+		}
+
+	case RHYTHMDB_QUERY_MODEL_LIMIT_TIME:
+		{
+			gulong limit_time;
+			gulong current_time;
+			
+			limit_time = g_value_get_ulong (g_value_array_get_nth (model->priv->limit_value, 0));
+			current_time = model->priv->total_duration;
+
+			if (entry)
+				current_time += rhythmdb_entry_get_ulong (entry, RHYTHMDB_PROP_DURATION);
+
+			result = (current_time <= limit_time);
+			break;
+		}
+	}
+
+	return result;
+}
+
+/* This should really be standard. */
+#define ENUM_ENTRY(NAME, DESC) { NAME, "" #NAME "", DESC }
+
+GType
+rhythmdb_query_model_limit_type_get_type (void)
+{
+	static GType etype = 0;
+
+	if (etype == 0)
+	{
+		static const GEnumValue values[] =
+		{
+
+			ENUM_ENTRY (RHYTHMDB_QUERY_MODEL_LIMIT_NONE, "No limit"),
+			ENUM_ENTRY (RHYTHMDB_QUERY_MODEL_LIMIT_COUNT, "Limit by number of entries (count)"),
+			ENUM_ENTRY (RHYTHMDB_QUERY_MODEL_LIMIT_SIZE, "Limit by data size (Mb)"),
+			ENUM_ENTRY (RHYTHMDB_QUERY_MODEL_LIMIT_TIME, "Limit by duration (seconds)"),
+			{ 0, 0, 0 }
+		};
+
+		etype = g_enum_register_static ("RhythmDBQueryModelLimitType", values);
+	}
+
+	return etype;
+}
+
+static gboolean
+rhythmdb_query_model_reapply_query_cb (RhythmDBQueryModel *model)
+{
+	rhythmdb_query_model_reapply_query (model, FALSE);
+	rhythmdb_do_full_query_async_parsed (model->priv->db,
+					     RHYTHMDB_QUERY_RESULTS (model),
+					     model->priv->query);
+	return TRUE;
+}
