@@ -15,7 +15,7 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
  *
  */
 
@@ -25,8 +25,6 @@
 #include "rb-source.h"
 #include "rhythmdb.h"
 #include "rb-sourcelist.h"
-#include "rb-library-source.h"
-#include "rb-iradio-source.h"
 
 G_BEGIN_DECLS
 
@@ -68,25 +66,27 @@ typedef struct
 	void	(*load_finish) (RBPlaylistManager *manager);
 } RBPlaylistManagerClass;
 
+typedef enum
+{
+	RB_PLAYLIST_EXPORT_TYPE_UNKNOWN,
+	RB_PLAYLIST_EXPORT_TYPE_M3U,
+	RB_PLAYLIST_EXPORT_TYPE_PLS,
+} RBPlaylistExportType;
+
 GType			rb_playlist_manager_get_type	(void);
 
 RBPlaylistManager *	rb_playlist_manager_new		(RBShell *shell,
-							 RBSourceList *sourcelist,
-							 RBLibrarySource *libsource,
-							 RBIRadioSource *iradio_source);
+							 RBSourceList *sourcelist);
 
 void			rb_playlist_manager_shutdown	(RBPlaylistManager *mgr);
 gboolean 		rb_playlist_manager_parse_file	(RBPlaylistManager *mgr,
 							 const char *uri,
 							 GError **error);
 
-void			rb_playlist_manager_set_source	(RBPlaylistManager *mgr,
-							 RBSource *player);
-
 void			rb_playlist_manager_load_playlists (RBPlaylistManager *mgr);
 
-void			rb_playlist_manager_save_playlists_async (RBPlaylistManager *mgr, gboolean force);
-void			rb_playlist_manager_save_playlists (RBPlaylistManager *mgr, gboolean force);
+gboolean		rb_playlist_manager_save_playlists (RBPlaylistManager *mgr, 
+							    gboolean force);
 
 RBSource *		rb_playlist_manager_new_playlist (RBPlaylistManager *mgr,
 							  const char *suggested_name,
