@@ -16,32 +16,54 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
  *
  */
 
-#include <config.h>
-#include <gtk/gtk.h>
-#include <libgnome/gnome-i18n.h>
-#include <libgnomevfs/gnome-vfs.h>
-#include <nautilus-burn-drive.h>
+#include "config.h"
+
 #include <string.h>
+#include <glib/gi18n.h>
+#include <gtk/gtk.h>
+
+#include <libgnomevfs/gnome-vfs.h>
+
+#include <nautilus-burn-drive.h>
+#ifndef NAUTILUS_BURN_CHECK_VERSION 	 
+#define NAUTILUS_BURN_CHECK_VERSION(a,b,c) FALSE 	 
+#endif
+
+#if NAUTILUS_BURN_CHECK_VERSION(2,15,3)
+#include <nautilus-burn.h>
+#endif
 
 #include "rb-removable-media-manager.h"
+#include "rb-library-source.h"
 #include "rb-sourcelist.h"
 #include "rb-removable-media-source.h"
 #include "rb-audiocd-source.h"
-#ifdef WITH_IPOD_SUPPORT
-#include "rb-ipod-source.h"
-#endif
 
+#include "rb-shell.h"
+#include "rb-shell-player.h"
 #include "rb-debug.h"
 #include "rb-dialog.h"
+#include "rb-stock-icons.h"
 #include "rhythmdb.h"
+#include "rb-marshal.h"
+#include "rb-util.h"
+
+#ifdef ENABLE_TRACK_TRANSFER
+#include "rb-encoder.h"
+#endif
+
+#ifndef HAVE_BURN_DRIVE_UNREF
+#define nautilus_burn_drive_unref nautilus_burn_drive_free
+#endif
 
 static void rb_removable_media_manager_class_init (RBRemovableMediaManagerClass *klass);
 static void rb_removable_media_manager_init (RBRemovableMediaManager *mgr);
 static void rb_removable_media_manager_dispose (GObject *object);
+static void rb_removable_media_manager_finalize (GObject *object);
 static void rb_removable_media_manager_set_property (GObject *object,
 					      guint prop_id,
 					      const GValue *value,
@@ -71,15 +93,13 @@ static void  rb_removable_media_manager_volume_mounted_cb (GnomeVFSVolumeMonitor
 static void  rb_removable_media_manager_volume_unmounted_cb (GnomeVFSVolumeMonitor *monitor,
 				GnomeVFSVolume *volume, 
 				gpointer data);
-static void rb_removable_media_manager_scan (RBRemovableMediaManager *manager);
+static gboolean rb_removable_media_manager_load_media (RBRemovableMediaManager *manager);
 
-typedef struct
-{
-	gboolean removed;
-	gboolean tray_opened;
-	RBRemovableMediaManager *manager;
-	NautilusBurnDrive *drive;
-} RbCdDriveInfo;
+#ifdef ENABLE_TRACK_TRANSFER
+static void do_transfer (RBRemovableMediaManager *manager);
+#endif
+static void rb_removable_media_manager_cmd_copy_tracks (GtkAction *action,
+							RBRemovableMediaManager *mgr);
 
 typedef struct
 {
@@ -94,8 +114,19 @@ typedef struct
 
 	GList *sources;
 	GHashTable *volume_mapping;
+#if !NAUTILUS_BURN_CHECK_VERSION(2,15,3)
 	GHashTable *cd_drive_mapping;
+#endif
 	GList *cur_volume_list;
+	gboolean scanned;
+
+	char *playing_uri;
+
+	GAsyncQueue *transfer_queue;
+	gboolean transfer_running;
+	gint transfer_total;
+	gint transfer_done;
+	double transfer_fraction;
 } RBRemovableMediaManagerPrivate;
 
 G_DEFINE_TYPE (RBRemovableMediaManager, rb_removable_media_manager, G_TYPE_OBJECT)
@@ -107,11 +138,14 @@ enum
 	PROP_SHELL,
 	PROP_SOURCELIST,
 	PROP_SOURCE,
+	PROP_SCANNED
 };
 
 enum
 {
 	MEDIUM_ADDED,
+	TRANSFER_PROGRESS,
+	CREATE_SOURCE,
 	LAST_SIGNAL
 };
 
@@ -119,9 +153,12 @@ static guint rb_removable_media_manager_signals[LAST_SIGNAL] = { 0 };
 
 static GtkActionEntry rb_removable_media_manager_actions [] =
 {
-	{ "RemovableSourceEject", NULL, N_("_Eject"), NULL,
+	{ "RemovableSourceEject", GNOME_MEDIA_EJECT, N_("_Eject"), NULL,
 	  N_("Eject this medium"),
 	  G_CALLBACK (rb_removable_media_manager_cmd_eject_medium) },
+	{ "RemovableSourceCopyAllTracks", GTK_STOCK_CDROM, N_("_Copy to library"), NULL,
+	  N_("Copy all tracks to the library"),
+	  G_CALLBACK (rb_removable_media_manager_cmd_copy_tracks) },
 	{ "MusicScanMedia", NULL, N_("_Scan Removable Media"), NULL,
 	  N_("Scan for new Removable Media"),
 	  G_CALLBACK (rb_removable_media_manager_cmd_scan_media) },
@@ -135,7 +172,7 @@ rb_removable_media_manager_class_init (RBRemovableMediaManagerClass *klass)
 	GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
 	object_class->dispose = rb_removable_media_manager_dispose;
-
+	object_class->finalize = rb_removable_media_manager_finalize;
 	object_class->set_property = rb_removable_media_manager_set_property;
 	object_class->get_property = rb_removable_media_manager_get_property;
 
@@ -161,6 +198,13 @@ rb_removable_media_manager_class_init (RBRemovableMediaManagerClass *klass)
 							      "RBSourceList",
 							      RB_TYPE_SOURCELIST,
 							      G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
+	g_object_class_install_property (object_class,
+					 PROP_SCANNED,
+					 g_param_spec_boolean ("scanned",
+						 	       "scanned",
+							       "Whether a scan has been performed",
+							       FALSE,
+							       G_PARAM_READABLE));
 
 	rb_removable_media_manager_signals[MEDIUM_ADDED] =
 		g_signal_new ("medium_added",
@@ -172,6 +216,26 @@ rb_removable_media_manager_class_init (RBRemovableMediaManagerClass *klass)
 			      G_TYPE_NONE,
 			      1, G_TYPE_OBJECT);
 
+	rb_removable_media_manager_signals[TRANSFER_PROGRESS] =
+		g_signal_new ("transfer-progress",
+			      RB_TYPE_REMOVABLE_MEDIA_MANAGER,
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (RBRemovableMediaManagerClass, transfer_progress),
+			      NULL, NULL,
+			      rb_marshal_VOID__INT_INT_DOUBLE,
+			      G_TYPE_NONE,
+			      3, G_TYPE_INT, G_TYPE_INT, G_TYPE_DOUBLE);
+	
+	rb_removable_media_manager_signals[CREATE_SOURCE] =
+		g_signal_new ("create-source",
+			      RB_TYPE_REMOVABLE_MEDIA_MANAGER,
+			      G_SIGNAL_RUN_LAST,
+			      G_STRUCT_OFFSET (RBRemovableMediaManagerClass, create_source),
+			      rb_signal_accumulator_object_handled, NULL,
+			      rb_marshal_OBJECT__OBJECT,
+			      RB_TYPE_SOURCE,
+			      1, GNOME_VFS_TYPE_VOLUME);
+
 	g_type_class_add_private (klass, sizeof (RBRemovableMediaManagerPrivate));
 }
 
@@ -181,6 +245,9 @@ rb_removable_media_manager_init (RBRemovableMediaManager *mgr)
 	RBRemovableMediaManagerPrivate *priv = REMOVABLE_MEDIA_MANAGER_GET_PRIVATE (mgr);
 
 	priv->volume_mapping = g_hash_table_new (NULL, NULL);
+	priv->transfer_queue = g_async_queue_new ();
+
+	g_idle_add ((GSourceFunc)rb_removable_media_manager_load_media, mgr);
 }
 
 static void
@@ -205,20 +272,26 @@ rb_removable_media_manager_dispose (GObject *object)
 		g_list_free (priv->sources);
 		priv->sources = NULL;
 	}
-	if (priv->volume_mapping) {
-		g_hash_table_destroy (priv->volume_mapping);
-		priv->volume_mapping = NULL;
-	}
-	if (priv->cd_drive_mapping) {
-		g_hash_table_destroy (priv->cd_drive_mapping);
-		priv->cd_drive_mapping = NULL;
-	}
 
 	priv->disposed = TRUE;
 
 	G_OBJECT_CLASS (rb_removable_media_manager_parent_class)->dispose (object);
 }
 
+static void
+rb_removable_media_manager_finalize (GObject *object)
+{
+	RBRemovableMediaManagerPrivate *priv = REMOVABLE_MEDIA_MANAGER_GET_PRIVATE (object);
+
+	g_hash_table_destroy (priv->volume_mapping);
+#if !NAUTILUS_BURN_CHECK_VERSION(2,15,3)
+	g_hash_table_destroy (priv->cd_drive_mapping);
+#endif
+	g_async_queue_unref (priv->transfer_queue);
+	
+	G_OBJECT_CLASS (rb_removable_media_manager_parent_class)->finalize (object);
+}
+
 static void
 rb_removable_media_manager_set_property (GObject *object,
 				  guint prop_id,
@@ -273,6 +346,9 @@ rb_removable_media_manager_get_property (GObject *object,
 	case PROP_SOURCELIST:
 		g_value_set_object (value, priv->sourcelist);
 		break;
+	case PROP_SCANNED:
+		g_value_set_boolean (value, priv->scanned);
+		break;
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
 		break;
@@ -289,6 +365,15 @@ rb_removable_media_manager_new (RBShell *shell,
 			     NULL);
 }
 
+#if !NAUTILUS_BURN_CHECK_VERSION(2,15,3)
+typedef struct
+{
+	gboolean removed;
+	gboolean tray_opened;
+	RBRemovableMediaManager *manager;
+	NautilusBurnDrive *drive;
+} RbCdDriveInfo;
+
 #ifdef HAVE_BURN_DRIVE_DOOR
 static
 gboolean poll_tray_opened (RbCdDriveInfo *info)
@@ -298,7 +383,7 @@ gboolean poll_tray_opened (RbCdDriveInfo *info)
 	GnomeVFSVolume *volume;
 
 	if (info->removed) {
-		nautilus_burn_drive_free (info->drive);
+		nautilus_burn_drive_unref (info->drive);
 		g_free (info);
 		return FALSE;
 	}
@@ -333,29 +418,137 @@ void begin_cd_drive_monitor (NautilusBurnDrive *drive, RBRemovableMediaManager *
 {
 #ifdef HAVE_BURN_DRIVE_DOOR
 	RBRemovableMediaManagerPrivate *priv = REMOVABLE_MEDIA_MANAGER_GET_PRIVATE (manager);
-	RbCdDriveInfo *info = g_new (RbCdDriveInfo, 1);
+	RbCdDriveInfo *info = g_new0 (RbCdDriveInfo, 1);
 	GnomeVFSVolumeMonitor *monitor=  gnome_vfs_get_volume_monitor ();
 	GnomeVFSVolume *volume;
 
 	info->drive = drive;
 	info->tray_opened = nautilus_burn_drive_door_is_open (drive);
 	info->manager = manager;
+
 	g_hash_table_insert (priv->cd_drive_mapping, drive, info);
 	g_timeout_add (1000, (GSourceFunc)poll_tray_opened, info);
-	if (!nautilus_burn_drive_door_is_open (drive)) {
-		volume = gnome_vfs_volume_monitor_get_volume_for_path (monitor, drive->device);
-		if (volume)
+
+	volume = gnome_vfs_volume_monitor_get_volume_for_path (monitor, drive->device);
+
+	if (volume) {
+		if (!nautilus_burn_drive_door_is_open (drive)) {
 			rb_removable_media_manager_mount_volume (manager, volume);
+		} else {
+			/* it may have got ejected while we weren't monitoring */
+			rb_removable_media_manager_unmount_volume (manager, volume);
+		}
 	}
 #endif
 }
 
-gboolean
+static NautilusBurnDrive *
+get_nautilus_burn_drive_for_path (const char *path)
+{
+#ifdef HAVE_BURN_DRIVE_NEW_FROM_PATH
+	return nautilus_burn_drive_new_from_path (path);
+#else
+	GList *drives, *l;
+	NautilusBurnDrive *path_drive = NULL;
+
+	drives = nautilus_burn_drive_get_list (FALSE, FALSE);
+	for (l = drives; l != NULL; l = g_list_next (l)) {
+		NautilusBurnDrive *drive = (NautilusBurnDrive*)l->data;
+
+		if (path_drive == NULL && strcmp (drive->device, path) == 0) {
+			path_drive = drive;
+		} else {
+			nautilus_burn_drive_unref (drive);
+		}
+	}
+	g_list_free (drives);
+
+	return path_drive;
+#endif
+}
+#endif /* NAUTILUS_BURN < 2.15.3 */
+
+static char *
+split_drive_from_cdda_uri (const char *uri)
+{
+	gchar *copy, *temp, *split;
+	int len;
+
+	if (!g_str_has_prefix (uri, "cdda://"))
+		return NULL;
+
+	len = strlen ("cdda://");
+	
+	copy = g_strdup (uri);
+	split = g_utf8_strrchr (copy + len, -1, ':');
+
+	if (split == NULL) {
+		/* invalid URI, it doesn't contain a ':' */
+		g_free (copy);
+		return NULL;
+	}
+
+	*split = 0;
+	temp = g_strdup (copy + len);
+	g_free (copy);
+
+	return temp;
+}
+
+static void
+rb_removable_media_manager_playing_uri_changed_cb (RBShellPlayer *player,
+						   const char *uri,
+						   RBRemovableMediaManager *manager)
+{
+	RBRemovableMediaManagerPrivate *priv = REMOVABLE_MEDIA_MANAGER_GET_PRIVATE (manager);
+	char *old_drive = NULL;
+	char *new_drive = NULL;
+
+	/* extract the drive paths */
+	if (priv->playing_uri)
+		old_drive = split_drive_from_cdda_uri (priv->playing_uri);
+
+	if (uri)
+		new_drive = split_drive_from_cdda_uri (uri);
+
+#if !NAUTILUS_BURN_CHECK_VERSION(2,15,3)
+	/* if the drive we're playing from has changed, adjust the polling */
+	if (old_drive == NULL || new_drive == NULL || strcmp (old_drive, new_drive) != 0) {
+		if (old_drive) {
+			NautilusBurnDrive *drive;
+
+			rb_debug ("restarting monitoring of drive %s after playing", old_drive);
+			drive = get_nautilus_burn_drive_for_path (old_drive);
+			begin_cd_drive_monitor (drive, manager);
+			nautilus_burn_drive_unref (drive);
+		}
+		
+		if (new_drive) {
+			NautilusBurnDrive *drive;
+
+			rb_debug ("stopping monitoring of drive %s while playing", new_drive);
+			drive = get_nautilus_burn_drive_for_path (new_drive);
+			/* removing it from the hash table makes it stop monitoring */
+			g_hash_table_remove (priv->cd_drive_mapping, drive);
+			nautilus_burn_drive_unref (drive);
+		}
+	}
+#endif
+
+	g_free (priv->playing_uri);
+	priv->playing_uri = (uri) ? g_strdup (uri) : NULL;
+}
+
+	
+static gboolean
 rb_removable_media_manager_load_media (RBRemovableMediaManager *manager)
 {
 	RBRemovableMediaManagerPrivate *priv = REMOVABLE_MEDIA_MANAGER_GET_PRIVATE (manager);
 	GnomeVFSVolumeMonitor *monitor = gnome_vfs_get_volume_monitor ();
+#if !NAUTILUS_BURN_CHECK_VERSION(2,15,3)
 	GList *drives;
+#endif
+	GObject *shell_player;
 
 	/*
 	 * Monitor new (un)mounted file systems to look for new media
@@ -381,13 +574,18 @@ rb_removable_media_manager_load_media (RBRemovableMediaManager *manager)
 	 * This needs to be done seperately from the above, because non-HAL systems don't
 	 * (currently) report audio cd insertions as mount events.
 	 */
+#if !NAUTILUS_BURN_CHECK_VERSION(2,15,3)
 	priv->cd_drive_mapping = g_hash_table_new_full (NULL, NULL, NULL, (GDestroyNotify)end_cd_drive_monitor);
 	drives = nautilus_burn_drive_get_list (FALSE, FALSE);
 	g_list_foreach (drives, (GFunc)begin_cd_drive_monitor, manager);
 	g_list_free (drives);
+#endif
 
-	/* scan for media */
-	rb_removable_media_manager_scan (manager);
+	/* monitor the playing song, to disable cd drive polling */
+	shell_player = rb_shell_get_player (priv->shell);
+	g_signal_connect (shell_player, "playing-uri-changed",
+			  G_CALLBACK (rb_removable_media_manager_playing_uri_changed_cb),
+			  manager);
 
 	return FALSE;
 }
@@ -403,6 +601,25 @@ rb_removable_media_manager_volume_mounted_cb (GnomeVFSVolumeMonitor *monitor,
 }
 
 
+static gboolean 
+remove_volume_by_source (GnomeVFSVolume *volume, RBSource *source, 
+			 RBSource *ref_source)
+{
+	return (ref_source == source);
+}
+
+static void
+rb_removable_media_manager_source_deleted_cb (RBSource *source, RBRemovableMediaManager *mgr)
+{
+	RBRemovableMediaManagerPrivate *priv = REMOVABLE_MEDIA_MANAGER_GET_PRIVATE (mgr);
+
+	rb_debug ("removing source %p", source);
+	g_hash_table_foreach_remove (priv->volume_mapping, 
+				     (GHRFunc)remove_volume_by_source,
+				     source);
+	priv->sources = g_list_remove (priv->sources, source);
+}
+
 static void
 rb_removable_media_manager_volume_unmounted_cb (GnomeVFSVolumeMonitor *monitor,
 			     GnomeVFSVolume *volume, 
@@ -414,13 +631,11 @@ rb_removable_media_manager_volume_unmounted_cb (GnomeVFSVolumeMonitor *monitor,
 	rb_removable_media_manager_unmount_volume (mgr, volume);
 }
 
-
 static void
 rb_removable_media_manager_mount_volume (RBRemovableMediaManager *mgr, GnomeVFSVolume *volume)
 {
 	RBRemovableMediaManagerPrivate *priv = REMOVABLE_MEDIA_MANAGER_GET_PRIVATE (mgr);
 	RBRemovableMediaSource *source = NULL;
-	RBShell *shell;
 	char *fs_type, *device_path, *display_name, *hal_udi, *icon_name;
 	GnomeVFSDeviceType device_type;
 
@@ -429,6 +644,9 @@ rb_removable_media_manager_mount_volume (RBRemovableMediaManager *mgr, GnomeVFSV
 	if (g_hash_table_lookup (priv->volume_mapping, volume) != NULL)
 		return;
 
+	if (!gnome_vfs_volume_is_mounted (volume))
+		return;
+
 	/* ignore network volumes */
 	device_type = gnome_vfs_volume_get_device_type (volume);
 	if (device_type == GNOME_VFS_DEVICE_TYPE_NFS ||
@@ -437,15 +655,13 @@ rb_removable_media_manager_mount_volume (RBRemovableMediaManager *mgr, GnomeVFSV
 	    device_type == GNOME_VFS_DEVICE_TYPE_NETWORK)
 		return;
 
-	g_object_get (G_OBJECT (mgr), "shell", &shell, NULL);
-
 	fs_type = gnome_vfs_volume_get_filesystem_type (volume);
 	device_path = gnome_vfs_volume_get_device_path (volume);
 	display_name = gnome_vfs_volume_get_display_name (volume);
 	hal_udi = gnome_vfs_volume_get_hal_udi (volume);
 	icon_name = gnome_vfs_volume_get_icon (volume);
-	rb_debug ("detecting new media - device_type=%d", device_type);
-	rb_debug ("detecting new media - volumd_type=%d", gnome_vfs_volume_get_volume_type (volume));
+	rb_debug ("detecting new media - device type=%d", device_type);
+	rb_debug ("detecting new media - volume type=%d", gnome_vfs_volume_get_volume_type (volume));
 	rb_debug ("detecting new media - fs type=%s", fs_type);
 	rb_debug ("detecting new media - device path=%s", device_path);
 	rb_debug ("detecting new media - display name=%s", display_name);
@@ -458,25 +674,25 @@ rb_removable_media_manager_mount_volume (RBRemovableMediaManager *mgr, GnomeVFSV
 	 * When volume is of the appropriate type, it creates a new source
 	 * to handle this volume
 	 */
+
+	g_signal_emit (G_OBJECT (mgr), rb_removable_media_manager_signals[CREATE_SOURCE], 0,
+		       volume, &source);
+
 	if (source == NULL && rb_audiocd_is_volume_audiocd (volume))
-		source = rb_audiocd_source_new (shell, volume);
-#ifdef WITH_IPOD_SUPPORT
-	if (source == NULL && rb_ipod_is_volume_ipod (volume))
-		source = rb_ipod_source_new (shell, volume);
-#endif
+		source = rb_audiocd_source_new (priv->shell, volume);
 
 	if (source) {
 		g_hash_table_insert (priv->volume_mapping, volume, source);
 		rb_removable_media_manager_append_media_source (mgr, source);
-	} else
-		rb_debug ("Unhanded media");
+	} else {
+		rb_debug ("Unhandled media");
+	}
 
 	g_free (fs_type);
 	g_free (device_path);
 	g_free (display_name);
 	g_free (hal_udi);
 	g_free (icon_name);
-	g_object_unref (G_OBJECT (shell));
 }
 
 static void
@@ -491,18 +707,9 @@ rb_removable_media_manager_unmount_volume (RBRemovableMediaManager *mgr, GnomeVF
 	source = g_hash_table_lookup (priv->volume_mapping, volume);
 	if (source) {
 		rb_source_delete_thyself (RB_SOURCE (source));
-		g_hash_table_remove (priv->volume_mapping, volume);
 	}
 }
 
-static void
-rb_removable_media_manager_source_deleted_cb (RBSource *source, RBRemovableMediaManager *mgr)
-{
-	RBRemovableMediaManagerPrivate *priv = REMOVABLE_MEDIA_MANAGER_GET_PRIVATE (mgr);
-
-	rb_debug ("removing source %p", source);
-	priv->sources = g_list_remove (priv->sources, source);
-}
 
 static void
 rb_removable_media_manager_append_media_source (RBRemovableMediaManager *mgr, RBRemovableMediaSource *source)
@@ -512,6 +719,7 @@ rb_removable_media_manager_append_media_source (RBRemovableMediaManager *mgr, RB
 	priv->sources = g_list_prepend (priv->sources, source);
 	g_signal_connect_object (G_OBJECT (source), "deleted",
 				 G_CALLBACK (rb_removable_media_manager_source_deleted_cb), mgr, 0);
+
 	g_signal_emit (G_OBJECT (mgr), rb_removable_media_manager_signals[MEDIUM_ADDED], 0,
 		       source);
 }
@@ -543,6 +751,15 @@ rb_removable_media_manager_set_uimanager (RBRemovableMediaManager *mgr,
 					      mgr);
 	}
 
+#ifndef ENABLE_TRACK_TRANSFER
+	{
+		GtkAction *action;
+
+		action = gtk_action_group_get_action (priv->actiongroup, "RemovableSourceCopyAllTracks");
+		gtk_action_set_visible (action, FALSE);
+	}
+#endif
+
 	gtk_ui_manager_insert_action_group (priv->uimanager,
 					    priv->actiongroup,
 					    0);
@@ -557,7 +774,7 @@ rb_removable_media_manager_eject_medium_cb (gboolean succeeded,
 	if (succeeded)
 		return;
 
-	rb_error_dialog (NULL, error, detailed_error);
+	rb_error_dialog (NULL, error, "%s", detailed_error);
 }
 
 static void
@@ -602,7 +819,17 @@ rb_removable_media_manager_unmount_volume_swap (GnomeVFSVolume *volume, RBRemova
 	rb_removable_media_manager_unmount_volume (manager, volume);
 }
 
-static void
+#if !NAUTILUS_BURN_CHECK_VERSION(2,15,3)
+static const char *
+nautilus_burn_drive_get_device (NautilusBurnDrive *drive)
+{
+	g_return_val_if_fail (drive != NULL, NULL);
+
+	return drive->device;
+}
+#endif
+
+void
 rb_removable_media_manager_scan (RBRemovableMediaManager *manager)
 {
 	RBRemovableMediaManagerPrivate *priv = REMOVABLE_MEDIA_MANAGER_GET_PRIVATE (manager);
@@ -610,6 +837,8 @@ rb_removable_media_manager_scan (RBRemovableMediaManager *manager)
 	GList *list, *it;
 	GnomeVFSVolume *volume;
 	struct VolumeCheckData check_data;
+
+	priv->scanned = TRUE;
 		
 	list = gnome_vfs_volume_monitor_get_mounted_volumes (monitor);
 	
@@ -634,10 +863,15 @@ rb_removable_media_manager_scan (RBRemovableMediaManager *manager)
 	g_list_free (list);
 
 	/* scan cd drives */
+#if NAUTILUS_BURN_CHECK_VERSION(2,15,3)
+	list = nautilus_burn_drive_monitor_get_drives (nautilus_burn_get_drive_monitor ());
+#else
 	list = nautilus_burn_drive_get_list (FALSE, FALSE);
+#endif
 	for  (it = list; it != NULL; it = g_list_next (it)) {
-		NautilusBurnDrive *drive = (NautilusBurnDrive*)it->data;
-		volume = gnome_vfs_volume_monitor_get_volume_for_path (monitor, drive->device);
+		NautilusBurnDrive *drive = (NautilusBurnDrive *)it->data;
+
+		volume = gnome_vfs_volume_monitor_get_volume_for_path (monitor, nautilus_burn_drive_get_device (drive));
 		if (volume) {
 			rb_removable_media_manager_mount_volume (manager, volume);
 			gnome_vfs_volume_unref (volume);
@@ -645,3 +879,169 @@ rb_removable_media_manager_scan (RBRemovableMediaManager *manager)
 	}
 	g_list_free (list);
 }
+
+
+#ifdef ENABLE_TRACK_TRANSFER
+/* Track transfer */
+
+typedef struct {
+	RBRemovableMediaManager *manager;
+	RhythmDBEntry *entry;
+	char *dest;
+	char *mime_type;
+	gboolean failed;
+	RBTranferCompleteCallback callback;
+	gpointer userdata;
+} TransferData;
+
+static void
+emit_progress (RBRemovableMediaManager *mgr)
+{
+	RBRemovableMediaManagerPrivate *priv = REMOVABLE_MEDIA_MANAGER_GET_PRIVATE (mgr);
+
+	g_signal_emit (G_OBJECT (mgr), rb_removable_media_manager_signals[TRANSFER_PROGRESS], 0,
+		       priv->transfer_done,
+		       priv->transfer_total,
+		       priv->transfer_fraction);
+}
+
+static void
+error_cb (RBEncoder *encoder, GError *error, TransferData *data)
+{
+	rb_debug ("Error transferring track to %s: %s", data->dest, error->message);
+	rb_error_dialog (NULL, _("Error transferring track"), "%s", error->message);
+
+	data->failed = TRUE;
+	rb_encoder_cancel (encoder);
+}
+
+static void
+progress_cb (RBEncoder *encoder, double fraction, TransferData *data)
+{
+	RBRemovableMediaManagerPrivate *priv = REMOVABLE_MEDIA_MANAGER_GET_PRIVATE (data->manager);
+
+	rb_debug ("transfer progress %f", (float)fraction);
+	priv->transfer_fraction = fraction;
+	emit_progress (data->manager);
+}
+
+static void
+completed_cb (RBEncoder *encoder, TransferData *data)
+{
+	RBRemovableMediaManagerPrivate *priv = REMOVABLE_MEDIA_MANAGER_GET_PRIVATE (data->manager);
+
+	rb_debug ("completed transferring track to %s", data->dest);
+	if (!data->failed)
+		(data->callback) (data->entry, data->dest, data->userdata);
+
+	priv->transfer_running = FALSE;
+	priv->transfer_done++;
+	priv->transfer_fraction = 0.0;
+	do_transfer (data->manager);
+
+	g_object_unref (G_OBJECT (encoder));
+	g_free (data->dest);
+	g_free (data->mime_type);
+	g_free (data);
+}
+
+static void
+do_transfer (RBRemovableMediaManager *manager)
+{
+	RBRemovableMediaManagerPrivate *priv = REMOVABLE_MEDIA_MANAGER_GET_PRIVATE (manager);
+	TransferData *data;
+	RBEncoder *encoder;
+
+	g_assert (rb_is_main_thread ());
+	
+	emit_progress (manager);
+
+	if (priv->transfer_running)
+		return;
+
+	data = g_async_queue_try_pop (priv->transfer_queue);
+	if (data == NULL) {
+		priv->transfer_total = 0;
+		priv->transfer_done = 0;
+		emit_progress (manager);
+		return;
+	}
+
+	priv->transfer_running = TRUE;
+	priv->transfer_fraction = 0.0;
+
+	encoder = rb_encoder_new ();
+	g_signal_connect (G_OBJECT (encoder),
+			  "error", G_CALLBACK (error_cb),
+			  data);
+	g_signal_connect (G_OBJECT (encoder),
+			  "progress", G_CALLBACK (progress_cb),
+			  data);
+	g_signal_connect (G_OBJECT (encoder),
+			  "completed", G_CALLBACK (completed_cb),
+			  data);
+	rb_encoder_encode (encoder, data->entry, data->dest, NULL);
+}
+
+void
+rb_removable_media_manager_queue_transfer (RBRemovableMediaManager *manager,
+					  RhythmDBEntry *entry,
+					  const char *dest,
+					  const char *mime_type,
+					  RBTranferCompleteCallback callback,
+					  gpointer userdata)
+{
+	RBRemovableMediaManagerPrivate *priv = REMOVABLE_MEDIA_MANAGER_GET_PRIVATE (manager);
+	TransferData *data;
+
+	g_assert (rb_is_main_thread ());
+
+	data = g_new0 (TransferData, 1);
+	data->manager = manager;
+	data->entry = entry;
+	data->dest = g_strdup (dest);
+	data->mime_type = g_strdup (mime_type);
+	data->callback = callback;
+	data->userdata = userdata;
+
+	g_async_queue_push (priv->transfer_queue, data);
+	priv->transfer_total++;
+	do_transfer (manager);
+}
+
+static gboolean
+copy_entry (RhythmDBQueryModel *model,
+	    GtkTreePath *path,
+	    GtkTreeIter *iter,
+	    GList **list)
+{
+	GList *l;
+	l = g_list_append (*list, rhythmdb_query_model_iter_to_entry (model, iter));
+	*list = l;
+	return FALSE;
+}
+#endif
+
+static void
+rb_removable_media_manager_cmd_copy_tracks (GtkAction *action, RBRemovableMediaManager *mgr)
+{
+#ifdef ENABLE_TRACK_TRANSFER
+	RBRemovableMediaManagerPrivate *priv = REMOVABLE_MEDIA_MANAGER_GET_PRIVATE (mgr);
+	RBRemovableMediaSource *source;
+	RBLibrarySource *library;
+	RhythmDBQueryModel *model;
+	GList *list = NULL;
+
+	source = RB_REMOVABLE_MEDIA_SOURCE (priv->selected_source);
+	g_object_get (G_OBJECT (source), "query-model", &model, NULL);
+	g_object_get (G_OBJECT (priv->shell), "library-source", &library, NULL);
+
+	gtk_tree_model_foreach (GTK_TREE_MODEL (model), (GtkTreeModelForeachFunc)copy_entry, &list);
+	rb_source_paste (RB_SOURCE (library), list);
+	g_list_free (list);
+
+	g_object_unref (G_OBJECT (model));
+	g_object_unref (G_OBJECT (library));
+#endif
+}
+
