@@ -15,13 +15,15 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
  *
  */
 
 #ifndef __RB_REMOVABLE_MEDIA_MANAGER_H
 #define __RB_REMOVABLE_MEDIA_MANAGER_H
 
+#include <libgnomevfs/gnome-vfs.h>
+
 #include "rb-source.h"
 #include "rhythmdb.h"
 #include "rb-shell.h"
@@ -58,13 +60,15 @@ typedef struct
 					 gint done,
 					 gint total,
 					 double fraction);
+	RBSource * (*create_source)	(RBRemovableMediaManager *mgr,
+					 GnomeVFSVolume *volume);
 } RBRemovableMediaManagerClass;
 
 
 RBRemovableMediaManager* rb_removable_media_manager_new		(RBShell *shell, RBSourceList *sourcelist);
 GType			rb_removable_media_manager_get_type	(void);
 
-gboolean		rb_removable_media_manager_load_media	(RBRemovableMediaManager *mgr);
+void			rb_removable_media_manager_scan (RBRemovableMediaManager *manager);
 
 #ifdef ENABLE_TRACK_TRANSFER
 void	rb_removable_media_manager_queue_transfer (RBRemovableMediaManager *mgr,
