@@ -16,7 +16,7 @@
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
- *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
+ *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301  USA.
  *
  */
 
@@ -51,7 +51,7 @@ static const GtkTargetEntry rb_entry_view_drag_types[] = {{  "text/uri-list", 0,
 struct RBEntryViewColumnSortData
 {
 	GCompareDataFunc func;
-	RhythmDBPropType prop_id;
+	gpointer data;
 };
 
 static void rb_entry_view_class_init (RBEntryViewClass *klass);
@@ -133,10 +133,12 @@ struct RBEntryViewPrivate
 	RhythmDBQueryModel *playing_model;
 	RhythmDBEntry *playing_entry;
 	gboolean playing_entry_in_view;
+	guint selection_changed_id;
 
 	gboolean is_drag_source;
 	gboolean is_drag_dest;
 
+	GdkPixbuf *playing_col_pixbuf;
 	GdkPixbuf *playing_pixbuf;
 	GdkPixbuf *paused_pixbuf;
 	GdkPixbuf *error_pixbuf;
@@ -147,7 +149,7 @@ struct RBEntryViewPrivate
 	gint sorting_order;
 	char *sorting_column_name;
 
-	gboolean have_selection;
+	gboolean have_selection, have_complete_selection;
 
 	GHashTable *column_key_map;
 
@@ -187,6 +189,8 @@ G_DEFINE_TYPE (RBEntryView, rb_entry_view, GTK_TYPE_SCROLLED_WINDOW)
 
 static guint rb_entry_view_signals[LAST_SIGNAL] = { 0 };
 
+static GQuark rb_entry_view_column_always_visible;
+
 static gboolean
 type_ahead_search_func (GtkTreeModel *model, gint column,
 			const gchar *key, GtkTreeIter *iter,
@@ -194,11 +198,13 @@ type_ahead_search_func (GtkTreeModel *model, gint column,
 {
 	RhythmDBEntry *entry;
 	gchar *folded;
+	const gchar *entry_folded;
 	gboolean res;
 
 	gtk_tree_model_get (model, iter, 0, &entry, -1);
 	folded = rb_search_fold (key);
-	res = (strstr (rb_refstring_get_folded (entry->title), folded) == NULL);
+	entry_folded = rhythmdb_entry_get_string (entry, RHYTHMDB_PROP_TITLE_FOLDED);
+	res = (strstr (entry_folded, folded) == NULL);
 	g_free (folded);
 	return res;
 }
@@ -323,9 +329,10 @@ rb_entry_view_class_init (RBEntryViewClass *klass)
 			      G_SIGNAL_RUN_LAST,
 			      G_STRUCT_OFFSET (RBEntryViewClass, show_popup),
 			      NULL, NULL,
-			      g_cclosure_marshal_VOID__VOID,
+			      g_cclosure_marshal_VOID__BOOLEAN,
 			      G_TYPE_NONE,
-			      0);
+			      1,
+			      G_TYPE_BOOLEAN);
 	rb_entry_view_signals[HAVE_SEL_CHANGED] =
 		g_signal_new ("have_selection_changed",
 			      G_OBJECT_CLASS_TYPE (object_class),
@@ -347,6 +354,8 @@ rb_entry_view_class_init (RBEntryViewClass *klass)
 			      0);
 
 	g_type_class_add_private (klass, sizeof (RBEntryViewPrivate));
+
+	rb_entry_view_column_always_visible = g_quark_from_static_string ("rb_entry_view_column_always_visible");
 }
 
 static void
@@ -358,13 +367,19 @@ rb_entry_view_init (RBEntryView *view)
 
 	icon_theme = gtk_icon_theme_get_default ();
 
-	view->priv->playing_pixbuf = gtk_icon_theme_load_icon (icon_theme,
+	view->priv->playing_col_pixbuf = gtk_icon_theme_load_icon (icon_theme,
                                    			       "stock_volume-max", 
                                    			       16,
                                    			       0, 
                                    			       NULL);
+
+	view->priv->playing_pixbuf = gtk_icon_theme_load_icon (icon_theme,
+                                   			       "stock_media-play", 
+                                   			       16,
+                                   			       0, 
+                                   			       NULL);
 	view->priv->paused_pixbuf = gtk_icon_theme_load_icon (icon_theme,
-                                   			      "stock_volume-0", 
+                                   			      "stock_media-pause", 
                                    			      16,
                                    			      0, 
                                    			      NULL);
@@ -400,9 +415,14 @@ rb_entry_view_finalize (GObject *object)
 	g_hash_table_destroy (view->priv->column_sort_data_map);
 	g_hash_table_destroy (view->priv->column_key_map);
 
-	g_object_unref (G_OBJECT (view->priv->playing_pixbuf));
-	g_object_unref (G_OBJECT (view->priv->paused_pixbuf));
-	g_object_unref (G_OBJECT (view->priv->error_pixbuf));
+	if (view->priv->playing_pixbuf)
+		g_object_unref (G_OBJECT (view->priv->playing_pixbuf));
+	if (view->priv->playing_col_pixbuf)
+		g_object_unref (G_OBJECT (view->priv->playing_col_pixbuf));
+	if (view->priv->paused_pixbuf)
+		g_object_unref (G_OBJECT (view->priv->paused_pixbuf));
+	if (view->priv->error_pixbuf)
+		g_object_unref (G_OBJECT (view->priv->error_pixbuf));
 
 	g_free (view->priv->sorting_key);
 
@@ -467,6 +487,7 @@ rb_entry_view_set_property (GObject *object,
 					 view,
 					 0);
 
+		gtk_tree_selection_unselect_all (view->priv->selection);
 		view->priv->model = new_model;
 		if (view->priv->sorting_column)
 			rb_entry_view_resort_model (view);
@@ -474,6 +495,7 @@ rb_entry_view_set_property (GObject *object,
 		gtk_tree_view_set_model (GTK_TREE_VIEW (view->priv->treeview),
 					 GTK_TREE_MODEL (new_model));
 		view->priv->have_selection = FALSE;
+		view->priv->have_complete_selection = FALSE;
 
 		g_signal_emit (G_OBJECT (view), rb_entry_view_signals[ENTRIES_REPLACED], 0);
 		break;
@@ -575,13 +597,13 @@ rb_entry_view_playing_cell_data_func (GtkTreeViewColumn *column, GtkCellRenderer
 				      RBEntryView *view)
 {
 	RhythmDBEntry *entry;
-	GdkPixbuf *pixbuf;
+	GdkPixbuf *pixbuf = NULL;
 
 	entry = rhythmdb_query_model_iter_to_entry (view->priv->model, iter);
 
 	if (entry == view->priv->playing_entry) {
 		switch (view->priv->playing_state) {
-		case RB_ENTRY_VIEW_PLAYING:	
+		case RB_ENTRY_VIEW_PLAYING:
 			pixbuf = view->priv->playing_pixbuf;
 			break;
 		case RB_ENTRY_VIEW_PAUSED:
@@ -591,10 +613,10 @@ rb_entry_view_playing_cell_data_func (GtkTreeViewColumn *column, GtkCellRenderer
 			pixbuf = NULL;
 			break;
 		}
-	} else if (entry->playback_error) {
+	}
+
+	if (pixbuf == NULL && rhythmdb_entry_get_string (entry, RHYTHMDB_PROP_PLAYBACK_ERROR)) {
 		pixbuf = view->priv->error_pixbuf;
-	} else {
-		pixbuf = NULL;
 	}
 
 	g_object_set (G_OBJECT (renderer), "pixbuf", pixbuf, NULL);
@@ -609,7 +631,9 @@ rb_entry_view_rating_cell_data_func (GtkTreeViewColumn *column, GtkCellRenderer
 
 	entry = rhythmdb_query_model_iter_to_entry (view->priv->model, iter);
 
-	g_object_set (G_OBJECT (renderer), "rating", entry->rating, NULL);
+	g_object_set (G_OBJECT (renderer),
+		      "rating", rhythmdb_entry_get_double (entry, RHYTHMDB_PROP_RATING),
+		      NULL);
 }
 
 static void
@@ -697,7 +721,6 @@ rb_entry_view_year_cell_data_func (GtkTreeViewColumn *column, GtkCellRenderer *r
 	}
 }
 
-#if 0
 static void
 rb_entry_view_quality_cell_data_func (GtkTreeViewColumn *column, GtkCellRenderer *renderer,
 				      GtkTreeModel *tree_model, GtkTreeIter *iter,
@@ -708,26 +731,34 @@ rb_entry_view_quality_cell_data_func (GtkTreeViewColumn *column, GtkCellRenderer
 
 	entry = rhythmdb_query_model_iter_to_entry (data->view->priv->model, iter);
 
-	bitrate = rhythmdb_entry_get_int (data->view->priv->db, entry,
-					  data->propid);
+	bitrate = rhythmdb_entry_get_ulong (entry, data->propid);
 
-	if (bitrate <= 0) {
-		g_object_set (G_OBJECT (renderer), "text", _("Unknown"), NULL);
-	} else if (bitrate <= 80) {
-		g_object_set (G_OBJECT (renderer), "text", _("Very Low"), NULL);
-	} else if (bitrate <= 112) {
-		g_object_set (G_OBJECT (renderer), "text", _("Low"), NULL);
-	} else if (bitrate <= 160) {
-		g_object_set (G_OBJECT (renderer), "text", _("Regular"), NULL);
-	} else if (bitrate <= 224) {
-		g_object_set (G_OBJECT (renderer), "text", _("High"), NULL);
-	} else if (bitrate <= 1410) {
-		g_object_set (G_OBJECT (renderer), "text", _("Very High"), NULL);
+	if (bitrate > 0) {
+		char *s = g_strdup_printf (_("%u kbps"), (guint)bitrate);
+		g_object_set (G_OBJECT (renderer), "text", s, NULL);
+		g_free (s);
 	} else {
-		g_object_set (G_OBJECT (renderer), "text", _("Perfect"), NULL);
+		g_object_set (G_OBJECT (renderer), "text", _("Unknown"), NULL);
 	}
 }
-#endif
+
+static void
+rb_entry_view_location_cell_data_func (GtkTreeViewColumn *column, GtkCellRenderer *renderer,
+				       GtkTreeModel *tree_model, GtkTreeIter *iter,
+				       struct RBEntryViewCellDataFuncData *data)
+{
+	RhythmDBEntry *entry;
+	const char *location;
+	char *str;
+
+	entry = rhythmdb_query_model_iter_to_entry (data->view->priv->model, iter);
+
+	location = rhythmdb_entry_get_string (entry, data->propid);
+	str = gnome_vfs_unescape_string_for_display (location);
+
+	g_object_set (G_OBJECT (renderer), "text", str, NULL);
+	g_free (str);
+}
 
 static void
 rb_entry_view_string_cell_data_func (GtkTreeViewColumn *column, GtkCellRenderer *renderer,
@@ -878,7 +909,7 @@ rb_entry_view_column_clicked_cb (GtkTreeViewColumn *column, RBEntryView *view)
 }
 
 void
-rb_entry_view_append_column (RBEntryView *view, RBEntryViewColumn coltype)
+rb_entry_view_append_column (RBEntryView *view, RBEntryViewColumn coltype, gboolean always_visible)
 {
 	GtkTreeViewColumn *column;
 	GtkCellRenderer *renderer = NULL;
@@ -975,7 +1006,6 @@ rb_entry_view_append_column (RBEntryView *view, RBEntryViewColumn coltype)
 		strings[1] = "0000";
 		strings[2] = _("Unknown");
 		break;
-#if 0
 	case RB_ENTRY_VIEW_COL_QUALITY:
 		propid = RHYTHMDB_PROP_BITRATE;
 		cell_data->propid = propid;
@@ -985,9 +1015,9 @@ rb_entry_view_append_column (RBEntryView *view, RBEntryViewColumn coltype)
 		title = _("_Quality");
 		key = "Quality";
 		strings[0] = title;
-		strings[1] = "000";
+		strings[1] = _("000 kbps");
+		strings[2] = _("Unknown");
 		break;
-#endif
 	case RB_ENTRY_VIEW_COL_RATING:
 		propid = RHYTHMDB_PROP_RATING;
 		sort_func = (GCompareDataFunc) rhythmdb_query_model_double_ceiling_sort_func;
@@ -1046,6 +1076,35 @@ rb_entry_view_append_column (RBEntryView *view, RBEntryViewColumn coltype)
 		strings[0] = title;
 		strings[1] = rb_entry_view_get_time_date_column_sample ();
 		break;
+	case RB_ENTRY_VIEW_COL_LAST_SEEN:
+		propid = RHYTHMDB_PROP_LAST_SEEN;
+		cell_data->propid = RHYTHMDB_PROP_LAST_SEEN_STR;
+		cell_data_func = (GtkTreeCellDataFunc) rb_entry_view_string_cell_data_func;
+		sort_propid = RHYTHMDB_PROP_LAST_SEEN;
+		sort_func = (GCompareDataFunc) rhythmdb_query_model_ulong_sort_func;
+		title = _("Last _Seen");
+		key = "LastSeen";
+		strings[0] = title;
+		strings[1] = rb_entry_view_get_time_date_column_sample ();
+		break;
+	case RB_ENTRY_VIEW_COL_LOCATION:
+		propid = RHYTHMDB_PROP_LOCATION;
+		cell_data->propid = RHYTHMDB_PROP_LOCATION;
+		cell_data_func = (GtkTreeCellDataFunc) rb_entry_view_location_cell_data_func;
+		sort_propid = RHYTHMDB_PROP_LOCATION;
+		sort_func = (GCompareDataFunc) rhythmdb_query_model_location_sort_func;
+		title = _("L_ocation");
+		key = "Location";
+		ellipsize = TRUE;
+		break;
+	case RB_ENTRY_VIEW_COL_ERROR:
+		propid = RHYTHMDB_PROP_PLAYBACK_ERROR;
+		cell_data->propid = RHYTHMDB_PROP_PLAYBACK_ERROR;
+		cell_data_func = (GtkTreeCellDataFunc) rb_entry_view_string_cell_data_func;
+		title = _("Error");
+		key = "Error";
+		ellipsize = TRUE;
+		break;
 	default:
 		g_assert_not_reached ();
 		propid = -1;
@@ -1082,8 +1141,14 @@ rb_entry_view_append_column (RBEntryView *view, RBEntryViewColumn coltype)
 	gtk_tree_view_column_set_sizing (column, GTK_TREE_VIEW_COLUMN_FIXED);
 	gtk_tree_view_column_set_clickable (column, TRUE);
 
+	if (always_visible)
+		g_object_set_qdata (G_OBJECT (column), 
+				    rb_entry_view_column_always_visible, 
+				    GINT_TO_POINTER (1));
+
 	g_hash_table_insert (view->priv->propid_column_map, GINT_TO_POINTER (propid), column);
-	rb_entry_view_append_column_custom (view, column, title, key, sort_func, sort_propid);
+
+	rb_entry_view_append_column_custom (view, column, title, key, sort_func, GINT_TO_POINTER (sort_propid));
 }
 
 void
@@ -1092,7 +1157,7 @@ rb_entry_view_append_column_custom (RBEntryView *view,
 				    const char *title,
 				    const char *key,
 				    GCompareDataFunc sort_func,
-				    RhythmDBPropType sort_prop_id)
+				    gpointer data)
 {
 	struct RBEntryViewColumnSortData *sortdata;
 
@@ -1113,7 +1178,7 @@ rb_entry_view_append_column_custom (RBEntryView *view,
 	if (sort_func != NULL) {
 		sortdata = g_new (struct RBEntryViewColumnSortData, 1);
 		sortdata->func = (GCompareDataFunc) sort_func;
-		sortdata->prop_id = sort_prop_id;
+		sortdata->data = data;
 		g_hash_table_insert (view->priv->column_sort_data_map, column, sortdata);
 	}
 	g_hash_table_insert (view->priv->column_key_map, g_strdup (key), column);
@@ -1186,7 +1251,7 @@ rb_entry_view_constructor (GType type, guint n_construct_properties,
 						     GDK_BUTTON1_MASK,
 						     rb_entry_view_drag_types,
 						     G_N_ELEMENTS (rb_entry_view_drag_types),
-						     GDK_ACTION_COPY | GDK_ACTION_MOVE);
+						     GDK_ACTION_COPY);
 	if (view->priv->is_drag_dest)
 		rb_tree_dnd_add_drag_dest_support (GTK_TREE_VIEW (view->priv->treeview),
 						   RB_TREE_DEST_CAN_DROP_BETWEEN | RB_TREE_DEST_EMPTY_VIEW_DROP,
@@ -1215,7 +1280,7 @@ rb_entry_view_constructor (GType type, guint n_construct_properties,
 							 view,
 							 NULL);
 
-		image_widget = gtk_image_new_from_pixbuf (view->priv->playing_pixbuf);
+		image_widget = gtk_image_new_from_pixbuf (view->priv->playing_col_pixbuf);
 		g_object_ref (G_OBJECT (image_widget));
 		gtk_object_sink (GTK_OBJECT (image_widget));
 		gtk_tree_view_column_set_widget (column, image_widget);
@@ -1291,6 +1356,7 @@ rb_entry_view_pixbuf_clicked_cb (RBEntryView          *view,
 {
 	GtkTreePath *path;
 	RhythmDBEntry *entry;
+	const gchar *error;
 
 	g_return_if_fail (path_string != NULL);
 
@@ -1299,9 +1365,9 @@ rb_entry_view_pixbuf_clicked_cb (RBEntryView          *view,
 
 	gtk_tree_path_free (path);
 
-	if (entry->playback_error) {
-		rb_error_dialog (NULL, _("Playback Error"),
-				 "%s", entry->playback_error);
+	error = rhythmdb_entry_get_string (entry, RHYTHMDB_PROP_PLAYBACK_ERROR);
+	if (error) {
+		rb_error_dialog (NULL, _("Playback Error"), "%s", error);
 	}
 }
 
@@ -1315,30 +1381,30 @@ rb_entry_view_playing_song_changed (RBShellPlayer *player,
   
  	g_return_if_fail (RB_IS_ENTRY_VIEW (view));
   
- 	if (view->priv->playing_state == RB_ENTRY_VIEW_NOT_PLAYING)
- 		return;
-  
- 	if (view->priv->playing_entry != NULL) {
- 		rb_entry_view_emit_row_changed (view, view->priv->playing_entry);
- 		g_object_unref (G_OBJECT (view->priv->playing_model));
- 	}
+	if (view->priv->playing_entry != NULL) {
+		if (view->priv->playing_state != RB_ENTRY_VIEW_NOT_PLAYING)
+			rb_entry_view_emit_row_changed (view, view->priv->playing_entry);
+		g_object_unref (G_OBJECT (view->priv->playing_model));
+	}
  	
  	view->priv->playing_entry = entry;
  	g_object_ref (G_OBJECT (view->priv->model));
  	view->priv->playing_model = view->priv->model;
  
- 	if (view->priv->playing_entry != NULL) {
- 		view->priv->playing_entry_in_view = 
- 			rb_entry_view_emit_row_changed (view, view->priv->playing_entry);
- 	}
- 
- 	if (view->priv->playing_entry
- 	    && view->priv->playing_entry_in_view) {
- 	    rb_entry_view_entry_is_visible (view, view->priv->playing_entry,
- 					    &realized, &visible, &iter);
- 	    if (realized && !visible)
- 		    rb_entry_view_scroll_to_iter (view, &iter);
-  	}
+ 	if (view->priv->playing_state != RB_ENTRY_VIEW_NOT_PLAYING) {
+		if (view->priv->playing_entry != NULL) {
+			view->priv->playing_entry_in_view = 
+				rb_entry_view_emit_row_changed (view, view->priv->playing_entry);
+		}
+	 
+		if (view->priv->playing_entry
+		    && view->priv->playing_entry_in_view) {
+		    rb_entry_view_entry_is_visible (view, view->priv->playing_entry,
+						    &realized, &visible, &iter);
+		    if (realized && !visible)
+			    rb_entry_view_scroll_to_iter (view, &iter);
+		}
+	}
 }
 
 static gboolean
@@ -1366,6 +1432,7 @@ rb_entry_view_get_selected_entries (RBEntryView *view)
 					     (GtkTreeSelectionForeachFunc) harvest_entries,
 					     (gpointer) &list);
 
+	list = g_list_reverse (list);
 	return list;
 }
 
@@ -1390,9 +1457,9 @@ rb_entry_view_button_press_cb (GtkTreeView *treeview,
 
 			g_list_free (selected);
 			
-			g_signal_emit (G_OBJECT (view), rb_entry_view_signals[SHOW_POPUP], 0);
 		}
-		return view->priv->have_selection;
+		g_signal_emit (G_OBJECT (view), rb_entry_view_signals[SHOW_POPUP], 0, (path != NULL));
+		return TRUE;
 	}
 
 	return FALSE;
@@ -1409,25 +1476,37 @@ rb_entry_view_popup_menu_cb (GtkTreeView *treeview,
 	return TRUE;
 }
 
-static void
-rb_entry_view_selection_changed_cb (GtkTreeSelection *selection,
-				   RBEntryView *view)
+static gboolean
+rb_entry_view_emit_selection_changed (RBEntryView *view)
 {
 	gboolean available;
-	GList *sel;
+	gint sel_count;
 
-	sel = rb_entry_view_get_selected_entries (view);
-	available = (sel != NULL);
-	g_list_free (sel);
+	sel_count = gtk_tree_selection_count_selected_rows (view->priv->selection);
+	available = (sel_count > 0);
 
 	if (available != view->priv->have_selection) {
+		gint entry_count;
+
+		entry_count = gtk_tree_model_iter_n_children (GTK_TREE_MODEL (view->priv->model), NULL);
+		view->priv->have_complete_selection = (sel_count == entry_count);
+
 		view->priv->have_selection = available;
 
 		g_signal_emit (G_OBJECT (view), rb_entry_view_signals[HAVE_SEL_CHANGED], 0, available);
 	}
 
+	view->priv->selection_changed_id = 0;
 	g_signal_emit (G_OBJECT (view), rb_entry_view_signals[SELECTION_CHANGED], 0);
+	return FALSE;
+}
 
+static void
+rb_entry_view_selection_changed_cb (GtkTreeSelection *selection,
+				    RBEntryView *view)
+{
+	if (view->priv->selection_changed_id == 0)
+		view->priv->selection_changed_id = g_idle_add ((GSourceFunc)rb_entry_view_emit_selection_changed, view);
 }
 
 gboolean
@@ -1436,6 +1515,12 @@ rb_entry_view_have_selection (RBEntryView *view)
 	return view->priv->have_selection;
 }
 
+gboolean
+rb_entry_view_have_complete_selection (RBEntryView *view)
+{
+	return view->priv->have_complete_selection;
+}
+
 static void
 rb_entry_view_row_activated_cb (GtkTreeView *treeview,
 			       GtkTreePath *path,
@@ -1485,6 +1570,8 @@ rb_entry_view_rows_reordered_cb (GtkTreeModel *model,
 	GList *selected_rows;
 	GList *i;
 	gint model_size;
+	gboolean scrolled = FALSE;
+	
 	rb_debug ("rows reordered");
 
 	model_size = gtk_tree_model_iter_n_children (model, NULL);
@@ -1506,6 +1593,14 @@ rb_entry_view_rows_reordered_cb (GtkTreeModel *model,
 				if (order[newindex] == index) {
 					newpath = gtk_tree_path_new_from_indices (newindex, -1);
 					gtk_tree_selection_select_path (view->priv->selection, newpath);
+					if (!scrolled) {
+						GtkTreeViewColumn *col;
+						GtkTreeView *treeview = GTK_TREE_VIEW (view->priv->treeview);
+
+						col = gtk_tree_view_get_column (treeview, 0);
+						gtk_tree_view_scroll_to_cell (treeview, newpath, col, TRUE, 0.5, 0.0);
+						scrolled = TRUE;
+					}
 					gtk_tree_path_free (newpath);
 					break;
 				}
@@ -1516,6 +1611,8 @@ rb_entry_view_rows_reordered_cb (GtkTreeModel *model,
 
 	g_list_foreach (selected_rows, (GFunc) gtk_tree_path_free, NULL);
 	g_list_free (selected_rows);
+
+	gtk_widget_queue_draw (GTK_WIDGET (view));
 }
 
 void
@@ -1683,7 +1780,7 @@ rb_entry_view_columns_config_changed_cb (GConfClient* client,
 static gint
 propid_from_name (const char *name)
 {
-	GEnumClass *prop_class = g_type_class_ref (RHYTHMDB_TYPE_PROP);
+	GEnumClass *prop_class = g_type_class_ref (RHYTHMDB_TYPE_PROP_TYPE);
 	GEnumValue *ev;
 	int ret;
 
@@ -1700,11 +1797,12 @@ set_column_visibility (guint propid, GtkTreeViewColumn *column, GList *visible_p
 {
 	gboolean visible;
 
-	/* title is always visible */
-	if (propid == RHYTHMDB_PROP_TITLE)
-		return;
+	if (g_object_get_qdata (G_OBJECT (column), 
+				rb_entry_view_column_always_visible) == GINT_TO_POINTER (1))
+		visible = TRUE;
+	else
+		visible = (g_list_find (visible_props, GINT_TO_POINTER (propid)) != NULL);
 
-	visible = (g_list_find (visible_props, GINT_TO_POINTER (propid)) != NULL);
 	gtk_tree_view_column_set_visible (column, visible);
 }
 
@@ -1841,6 +1939,61 @@ rb_entry_view_resort_model (RBEntryView *view)
 
 	rhythmdb_query_model_set_sort_order (view->priv->model,
 					     sort_data->func,
-					     sort_data->prop_id,
+					     sort_data->data,
+					     NULL,
 					     (view->priv->sorting_order == GTK_SORT_DESCENDING));
 }
+
+/* This should really be standard. */
+#define ENUM_ENTRY(NAME, DESC) { NAME, "" #NAME "", DESC }
+
+GType
+rb_entry_view_column_get_type (void)
+{
+	static GType etype = 0;
+
+	if (etype == 0)	{
+		static const GEnumValue values[] = {
+			ENUM_ENTRY (RB_ENTRY_VIEW_COL_TRACK_NUMBER, "Track Number"),
+			ENUM_ENTRY (RB_ENTRY_VIEW_COL_TITLE, "Title"),
+			ENUM_ENTRY (RB_ENTRY_VIEW_COL_ARTIST, "Artist"),
+			ENUM_ENTRY (RB_ENTRY_VIEW_COL_ALBUM, "Album"),
+			ENUM_ENTRY (RB_ENTRY_VIEW_COL_GENRE, "Genre"),
+			ENUM_ENTRY (RB_ENTRY_VIEW_COL_DURATION, "Duration"),
+			ENUM_ENTRY (RB_ENTRY_VIEW_COL_QUALITY, "Quality"),
+			ENUM_ENTRY (RB_ENTRY_VIEW_COL_RATING, "Rating"),
+			ENUM_ENTRY (RB_ENTRY_VIEW_COL_PLAY_COUNT, "Play Count"),
+			ENUM_ENTRY (RB_ENTRY_VIEW_COL_YEAR, "Year"),
+			ENUM_ENTRY (RB_ENTRY_VIEW_COL_LAST_PLAYED, "Last Played"),
+			ENUM_ENTRY (RB_ENTRY_VIEW_COL_FIRST_SEEN, "First Seen"),
+			ENUM_ENTRY (RB_ENTRY_VIEW_COL_LAST_SEEN, "Last Seen"),
+			ENUM_ENTRY (RB_ENTRY_VIEW_COL_LOCATION, "Location"),
+			ENUM_ENTRY (RB_ENTRY_VIEW_COL_ERROR, "Error"),
+			{ 0, 0, 0 }
+		};
+
+		etype = g_enum_register_static ("RBEntryViewColumn", values);
+	}
+
+	return etype;
+}
+
+GType
+rb_entry_view_state_get_type (void)
+{
+	static GType etype = 0;
+
+	if (etype == 0)	{
+		static const GEnumValue values[] = {
+			ENUM_ENTRY (RB_ENTRY_VIEW_NOT_PLAYING, "Not Playing"),
+			ENUM_ENTRY (RB_ENTRY_VIEW_PLAYING, "Playing"),
+			ENUM_ENTRY (RB_ENTRY_VIEW_PAUSED, "Paused"),
+			{ 0, 0, 0 }
+		};
+
+		etype = g_enum_register_static ("RBEntryViewState", values);
+	}
+
+	return etype;
+}
+
